# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

from collections import defaultdict
from typing import Optional, Union

import torch
from torch.cuda.amp.grad_scaler import GradScaler, OptState, _refresh_per_optimizer_state
from torch.optim import Optimizer

from composer.utils import dist

__all__ = ['ClosureGradScaler']


class ClosureGradScaler(GradScaler):
    """ClosureGradScaler allows for gradient scaling during with closures.

    We use closures with optimizers (see `here <https://pytorch.org/docs/stable/optim.html>`__)
    during training in order to support certain algorithms like
    :class:`~composer.algorithms.SAM`. This class allows us to perform gradient
    scaling (see `here <https://pytorch.org/docs/stable/amp.html#torch.cuda.amp.GradScaler>`__)
    along with the use of closures during training.

    Args:
        ddp_reduce_scalar_and (Callable[[bool], bool]): A function that performs a
            ddp reduction with an `and` operation. Used to determine whether
            or not to continue computing an optimizer's `step` based on the presence
            of `inf/nan` in the gradients.
        ddp_reduce_tensor_sum (Callable[[Tensor], Tensor]): A function that performs
            a ddp reduction across tensors with a `sum` operation. Used to aggregate
            `inf/nan` information stored in tensors across devices.
    """

    def _force_scaler_ready(self, optimizer: Optimizer):
        optimizer_state = self._per_optimizer_states[id(optimizer)]
        optimizer_state['stage'] = OptState.READY

    def _empty_all_grads(self, optimizer):
        for group in optimizer.param_groups:
            for param in group['params']:
                if param.grad is not None:
                    param.grad = None

    def _unscale_grads_and_continue(self, optimizer: Optimizer):
        if (not self._enabled):
            return True
        self._check_scale_growth_tracker('step')
        optimizer_state = self._per_optimizer_states[id(optimizer)]

        if optimizer_state['stage'] is OptState.STEPPED:
            raise RuntimeError('step() has already been called since the last update().')

        if optimizer_state['stage'] is OptState.READY:
            self.unscale_(optimizer)
        inf_detected = sum(v.item() for v in optimizer_state['found_inf_per_device'].values())
        return not inf_detected

    def step(self, optimizer: Optimizer, *args, **kwargs):
        """Step the optimizer with amp.

        Always called before the optimizer step. Checks if the optimizer can handle AMP closures (currently only
        Composer's SAM optimizer) If so, it passes an AMP-modified closure to the optimizer.
        """
        closure = kwargs['closure']

        def _amp_closure(**kwargs):
            self._force_scaler_ready(optimizer)
            self._empty_all_grads(optimizer)

            retval: float = closure(**kwargs)

            should_continue = self._unscale_grads_and_continue(optimizer)
            other_should_continue = dist.all_gather_object(should_continue)

            return retval if all(other_should_continue) else None

        return optimizer.step(closure=_amp_closure)  # type: ignore

    # Mostly copied from original grad_scaler implementation
    # See: https://pytorch.org/docs/stable/_modules/torch/cuda/amp/grad_scaler.html#GradScaler
    def update(self, new_scale: Optional[Union[float, torch.FloatTensor]] = None):
        """Updates the scale factor.

        If any optimizer steps were skipped, the scale is multiplied by ``backoff_factor``
        to reduce it. If ``growth_interval`` non-skipped iterations occurred consecutively,
        the scale is multiplied by ``growth_factor`` to increase it.

        Passing ``new_scale`` sets the new scale value manually. (``new_scale`` is not
        used directly; it is used to fill GradScaler's internal scale tensor. So, if
        ``new_scale`` was a tensor, later in-place changes to that tensor will not further
        affect the scale that the GradScaler uses internally.)

        .. warning::

            This method should only be called at the end of the iteration, after ``scaler.step(optimizer)`` has
            been invoked for all optimizers used this iteration.

        Args:
            new_scale (float | FloatTensor, optional):  New scale factor. (default: ``None``)
        """
        if not self._enabled:
            return

        _scale, _growth_tracker = self._check_scale_growth_tracker('update')

        if new_scale is not None:
            # Accept a new user-defined scale.
            if isinstance(new_scale, float):
                self._scale.fill_(new_scale)  # type: ignore[union-attr]
            else:
                reason = 'new_scale should be a float or a 1-element torch.cuda.FloatTensor with requires_grad=False.'
                assert isinstance(new_scale, torch.cuda.FloatTensor), reason  # type: ignore[attr-defined]
                assert new_scale.numel() == 1, reason
                assert new_scale.requires_grad is False, reason
                self._scale.copy_(new_scale)  # type: ignore[union-attr]
        else:
            # Consume shared inf/nan data collected from optimizers to update the scale.
            # If all found_inf tensors are on the same device as self._scale, this operation is asynchronous.
            found_infs = [
                found_inf.to(device=_scale.device, non_blocking=True)
                for state in self._per_optimizer_states.values()
                for found_inf in state['found_inf_per_device'].values()
            ]

            assert len(found_infs) > 0, 'No inf checks were recorded prior to update.'

            found_inf_combined = found_infs[0]
            if len(found_infs) > 1:
                for i in range(1, len(found_infs)):
                    found_inf_combined += found_infs[i]

            # This is the only line changed from original grad_scaler implementation
            dist.all_reduce(found_inf_combined, reduce_operation='SUM')

            torch._amp_update_scale_(_scale, _growth_tracker, found_inf_combined, self._growth_factor,
                                     self._backoff_factor, self._growth_interval)

        # To prepare for next iteration, clear the data collected from optimizers this iteration.
        self._per_optimizer_states = defaultdict(_refresh_per_optimizer_state)



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0
from typing import Callable, Iterable, Type, TypeVar, cast

import torch
import torchvision.transforms.functional
from PIL.Image import Image as PillowImage

_InputImgT = TypeVar('_InputImgT', torch.Tensor, PillowImage)
_OutputImgT = TypeVar('_OutputImgT', torch.Tensor, PillowImage)


def image_as_type(image: _InputImgT, typ: Type[_OutputImgT]) -> _OutputImgT:
    """Converts between :class:`torch.Tensor` and :class:`PIL.Image.Image` image representations.

    Args:
        image (torch.Tensor | PIL.Image.Image): A single image
            represented as a :class:`PIL.Image.Image` or
            a rank 2 or rank 3 :class:`torch.Tensor` in ``HW`` or ``CHW`` format.
            A rank 4 or higher tensor can also be provided as long as no type
            conversion is needed; in this case, the input tensor will be
            returned. This case is allowed so that functions that natively
            operate on batch tensors can safely call
            ``image_as_type(image, torch.Tensor)`` without additional error
            and type checking.
        typ (torch.Tensor | PIL.Image.Image): Type of the
            copied image. Must be :class:`PIL.Image.Image` or :class:`torch.Tensor`.

    Returns:
        A copy of ``image`` with type ``typ``.

    Raises:
        TypeError: if ``typ`` is not one of :class:`torch.Tensor` or
            :class:`PIL.Image.Image`.
        ValueError: if ``image`` cannot be converted to the ``typ``,
            such as when requesting conversion of a rank 4 tensor to
            :class:`PIL.Image.Image`.

    """
    if isinstance(image, typ):
        return image
    if not typ in (torch.Tensor, PillowImage):
        raise TypeError(f'Only typ={{torch.Tensor, Image}} is supported; got {typ}')

    if typ is torch.Tensor:
        return cast(_OutputImgT, torchvision.transforms.functional.to_tensor(image))  # PIL -> Tensor
    return cast(_OutputImgT, torchvision.transforms.functional.to_pil_image(image))  # Tensor -> PIL


def map_pillow_function(f_pil: Callable[[PillowImage], PillowImage], imgs: _OutputImgT) -> _OutputImgT:
    """Lifts a function that requires pillow images to also work on tensors.

    Args:
        f_pil ((PIL.Image.Image) -> PIL.Image.Image): A callable that takes maps :class:`PIL.Image.Image` objects.
            to other :class:`PIL.Image.Image` objects.
        imgs (torch.Tensor | PIL.Image.Image): a :class:`PIL.Image.Image` or a :class:`torch.Tensor` in ``HW``,
            ``CHW`` or ``NCHW`` format.

    Returns:
        The result of applying ``f_pil`` to each image in ``imgs``, converted
        back to the same type and (if applicable) tensor layout as ``imgs``.
    """
    single_image_input = not isinstance(imgs, Iterable)
    single_image_input |= isinstance(imgs, torch.Tensor) and imgs.ndim == 3
    imgs_as_iterable = [imgs] if single_image_input else imgs
    imgs_as_iterable = cast(type(imgs_as_iterable), imgs_as_iterable)

    imgs_pil = [image_as_type(img, PillowImage) for img in imgs_as_iterable]
    imgs_out_pil = [f_pil(img_pil) for img_pil in imgs_pil]
    imgs_out = [image_as_type(img_pil, type(imgs_as_iterable[0])) for img_pil in imgs_out_pil]

    if isinstance(imgs, torch.Tensor) and imgs.ndim == 4:  # batch of imgs
        imgs_out = [torch.unsqueeze(cast(torch.Tensor, img), 0) for img in imgs_out]
        imgs_out = torch.cat(imgs_out, dim=0)
    if single_image_input:
        imgs_out = imgs_out[0]
    imgs_out = cast(_OutputImgT, imgs_out)
    return imgs_out



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""ViT Small Patch 16 for image classification."""

from composer.models.vit_small_patch16.model import vit_small_patch16 as vit_small_patch16

__all__ = ['vit_small_patch16']

_task = 'Image Classification'
_dataset = 'ImageNet'
_name = 'ViT-Small-Patch16'
_quality = '74.52'
_metric = 'Top-1 Accuracy'
_ttt = '1d 59m'
_hparams = 'vit_small_patch16.yaml'



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Core CutMix classes and functions."""

from __future__ import annotations

import logging
from typing import Any, Callable, Optional, Tuple, Union

import numpy as np
import torch
from torch import Tensor

from composer.core import Algorithm, Event, State
from composer.loggers import Logger
from composer.loss.utils import ensure_targets_one_hot

log = logging.getLogger(__name__)

__all__ = ['CutMix', 'cutmix_batch']


def cutmix_batch(input: Tensor,
                 target: Tensor,
                 length: Optional[float] = None,
                 alpha: float = 1.,
                 bbox: Optional[Tuple] = None,
                 indices: Optional[torch.Tensor] = None,
                 uniform_sampling: bool = False) -> Tuple[torch.Tensor, torch.Tensor, float, Tuple]:
    """Create new samples using combinations of pairs of samples.

    This is done by masking a region of each image in ``input`` and filling
    the masked region with the corresponding content from a random different
    image in``input``.

    The position of the masked region is determined by drawing a center point
    uniformly at random from all spatial positions.

    The area of the masked region is computed using either ``length`` or
    ``alpha``. If ``length`` is provided, it directly determines the size
    of the masked region. If it is not provided, the fraction of the input
    area to mask is drawn from a ``Beta(alpha, alpha)`` distribution.
    The original paper uses a fixed value of ``alpha = 1``.

    Alternatively, one may provide a bounding box to mask directly, in
    which case ``alpha`` is ignored and ``length`` must not be provided.

    The same masked region is used for the whole batch.

    .. note::
        The masked region is clipped at the spatial boundaries of the inputs.
        This means that there is no padding required, but the actual region
        used may be smaller than the nominal size computed using ``length``
        or ``alpha``.

    Args:
        input (torch.Tensor): input tensor of shape ``(N, C, H, W)``.
        target (torch.Tensor): target tensor of either shape ``N`` or
            ``(N, num_classes)``. In the former case, elements of ``target``
            must be integer class ids in the range ``0..num_classes``. In the
            latter case, rows of ``target`` may be arbitrary vectors of targets,
            including, e.g., one-hot encoded class labels, smoothed class
            labels, or multi-output regression targets.
        length (float, optional): Relative side length of the masked region.
            If specified, ``length`` is interpreted as a fraction of ``H`` and
            ``W``, and the resulting box is of size ``(length * H, length * W)``.
            Default: ``None``.
        alpha (float, optional): parameter for the Beta distribution over
            the fraction of the input to mask. Ignored if ``length`` is
            provided. Default: ``1``.
        bbox (tuple, optional): predetermined ``(x1, y1, x2, y2)``
            coordinates of the bounding box. Default: ``None``.
        indices (torch.Tensor, optional): Permutation of the samples to use.
            Default: ``None``.
        uniform_sampling (bool, optional): If ``True``, sample the bounding box
            such that each pixel has an equal probability of being mixed.
            If ``False``, defaults to the sampling used in the original paper
            implementation. Default: ``False``.

    Returns:
        input_mixed (torch.Tensor): batch of inputs after cutmix has been
            applied.
        target_perm (torch.Tensor): The labels of the mixed-in examples
        area (float): The fractional area of the unmixed region.
        bounding_box (tuple): the ``(left, top, right, bottom)`` coordinates of
            the bounding box that defines the mixed region.

    Raises:
        ValueError: If both ``length`` and ``bbox`` are provided.

    Example:
        .. testcode::

            import torch
            from composer.functional import cutmix_batch

            N, C, H, W = 2, 3, 4, 5
            num_classes = 10
            X = torch.randn(N, C, H, W)
            y = torch.randint(num_classes, size=(N,))
            X_mixed, target_perm, area, _ = cutmix_batch(X, y, alpha=0.2)
    """
    if bbox is not None and length is not None:
        raise ValueError(f'Cannot provide both length and bbox; got {length} and {bbox}')

    # Create shuffled indicies across the batch in preparation for cutting and mixing.
    # Use given indices if there are any.
    if indices is None:
        shuffled_idx = _gen_indices(input)
    else:
        shuffled_idx = indices

    H, W = input.shape[-2], input.shape[-1]

    # figure out fraction of area to cut
    if length is None:
        cutmix_lambda = _gen_cutmix_coef(alpha)
    else:
        cut_w = int(length * W)
        cut_h = int(length * H)
        cutmix_lambda = (cut_w * cut_h) / (H * W)

    # Create the new inputs.
    X_cutmix = torch.clone(input)
    # Sample a rectangular box using lambda. Use variable names from the paper.
    if bbox:
        rx, ry, rw, rh = bbox[0], bbox[1], bbox[2], bbox[3]
        box_area = (rw - rx) * (rh - ry)
        cutmix_lambda = box_area / (H * W)
    else:
        rx, ry, rw, rh = _rand_bbox(input.shape[2], input.shape[3], cutmix_lambda, uniform_sampling=uniform_sampling)
        bbox = (rx, ry, rw, rh)

    # Fill in the box with a part of a random image.
    X_cutmix[:, :, rx:rw, ry:rh] = X_cutmix[shuffled_idx, :, rx:rw, ry:rh]
    # adjust lambda to exactly match pixel ratio. This is an implementation detail taken from
    # the original implementation, and implies lambda is not actually beta distributed.
    adjusted_lambda = _adjust_lambda(input, bbox)

    # Make a shuffled version of y for interpolation
    y_shuffled = target[shuffled_idx]

    return X_cutmix, y_shuffled, adjusted_lambda, bbox


class CutMix(Algorithm):
    """`CutMix <https://arxiv.org/abs/1905.04899>`_ trains the network on non-overlapping combinations
    of pairs of examples and interpolated targets rather than individual examples and targets.

    This is done by taking a non-overlapping combination of a given batch X with a
    randomly permuted copy of X. The area is drawn from a ``Beta(alpha, alpha)``
    distribution.

    Training in this fashion sometimes reduces generalization error.

    Args:
        alpha (float, optional): the psuedocount for the Beta distribution
            used to sample area parameters. As ``alpha`` grows, the two samples
            in each pair tend to be weighted more equally. As ``alpha``
            approaches 0 from above, the combination approaches only using
            one element of the pair. Default: ``1``.
        interpolate_loss (bool, optional): Interpolates the loss rather than the labels.
            A useful trick when using a cross entropy loss. Will produce incorrect behavior
            if the loss is not a linear function of the targets. Default: ``False``
        uniform_sampling (bool, optional): If ``True``, sample the bounding
            box such that each pixel has an equal probability of being mixed.
            If ``False``, defaults to the sampling used in the original
            paper implementation. Default: ``False``.
        input_key (str | int | Tuple[Callable, Callable] | Any, optional): A key that indexes to the input
            from the batch. Can also be a pair of get and set functions, where the getter
            is assumed to be first in the pair.  The default is 0, which corresponds to any sequence, where the first element
            is the input. Default: ``0``.
        target_key (str | int | Tuple[Callable, Callable] | Any, optional): A key that indexes to the target
            from the batch. Can also be a pair of get and set functions, where the getter
            is assumed to be first in the pair. The default is 1, which corresponds to any sequence, where the second element
            is the target. Default: ``1``.

    Example:
        .. testcode::

            from composer.algorithms import CutMix
            algorithm = CutMix(alpha=0.2)
            trainer = Trainer(
                model=model,
                train_dataloader=train_dataloader,
                eval_dataloader=eval_dataloader,
                max_duration="1ep",
                algorithms=[algorithm],
                optimizers=[optimizer]
            )
    """

    def __init__(
        self,
        alpha: float = 1.,
        interpolate_loss: bool = False,
        uniform_sampling: bool = False,
        input_key: Union[str, int, Tuple[Callable, Callable], Any] = 0,
        target_key: Union[str, int, Tuple[Callable, Callable], Any] = 1,
    ):
        self.alpha = alpha
        self.interpolate_loss = interpolate_loss
        self._uniform_sampling = uniform_sampling

        self._indices = torch.Tensor()
        self._cutmix_lambda = 0.0
        self._bbox: Tuple[int, int, int, int] = (0, 0, 0, 0)
        self._permuted_target = torch.Tensor()
        self._adjusted_lambda = 0.0
        self.input_key, self.target_key = input_key, target_key

    def match(self, event: Event, state: State) -> bool:
        if self.interpolate_loss:
            return event in [Event.BEFORE_FORWARD, Event.BEFORE_BACKWARD]
        else:
            return event in [Event.BEFORE_FORWARD, Event.BEFORE_LOSS]

    def apply(self, event: Event, state: State, logger: Logger) -> None:
        input = state.batch_get_item(key=self.input_key)
        target = state.batch_get_item(key=self.target_key)

        if not isinstance(input, torch.Tensor):
            raise NotImplementedError('Multiple tensors for inputs not supported yet.')
        if not isinstance(target, torch.Tensor):
            raise NotImplementedError('Multiple tensors for targets not supported yet.')

        alpha = self.alpha

        if event == Event.BEFORE_FORWARD:
            # these are saved only for testing
            self._indices = _gen_indices(input)

            _cutmix_lambda = _gen_cutmix_coef(alpha)
            self._bbox = _rand_bbox(input.shape[2],
                                    input.shape[3],
                                    _cutmix_lambda,
                                    uniform_sampling=self._uniform_sampling)
            self._adjusted_lambda = _adjust_lambda(input, self._bbox)

            new_input, self._permuted_target, _, _ = cutmix_batch(input=input,
                                                                  target=target,
                                                                  alpha=self.alpha,
                                                                  bbox=self._bbox,
                                                                  indices=self._indices,
                                                                  uniform_sampling=self._uniform_sampling)

            state.batch_set_item(key=self.input_key, value=new_input)

        if not self.interpolate_loss and event == Event.BEFORE_LOSS:
            # Interpolate the targets
            if not isinstance(state.outputs, torch.Tensor):
                raise NotImplementedError('Multiple output tensors not supported yet')
            if not isinstance(target, torch.Tensor):
                raise NotImplementedError('Multiple target tensors not supported yet')
            if self._permuted_target.ndim > 2 and self._permuted_target.shape[-2:] == input.shape[-2:]:
                # Target has the same height and width as the input, no need to interpolate.
                x1, y1, x2, y2 = self._bbox
                target[..., x1:x2, y1:y2] = self._permuted_target[..., x1:x2, y1:y2]
            else:
                # Need to interpolate on dense/one-hot targets.
                target = ensure_targets_one_hot(state.outputs, target)
                permuted_target = ensure_targets_one_hot(state.outputs, self._permuted_target)
                # Interpolate to get the new target
                target = self._adjusted_lambda * target + (1 - self._adjusted_lambda) * permuted_target
            # Create the new batch
            state.batch_set_item(key=self.target_key, value=target)

        if self.interpolate_loss and event == Event.BEFORE_BACKWARD:
            if self._permuted_target.ndim > 2 and self._permuted_target.shape[-2:] == input.shape[-2:]:
                raise ValueError("Can't interpolate loss when target has the same height and width as the input")

            # Grab the loss function
            if hasattr(state.model, 'loss'):
                loss_fn = state.model.loss
            elif hasattr(state.model, 'module') and hasattr(state.model.module, 'loss'):
                if isinstance(state.model.module, torch.nn.Module):
                    loss_fn = state.model.module.loss
                else:
                    raise TypeError('state.model.module must be a torch module')
            else:
                raise AttributeError('Loss must be accessible via model.loss or model.module.loss')
            # Verify that the loss is callable
            if not callable(loss_fn):
                raise TypeError('Loss must be callable')
            # Interpolate the loss
            new_loss = loss_fn(state.outputs, (input, self._permuted_target))
            if not isinstance(state.loss, torch.Tensor):
                raise NotImplementedError('Multiple losses not supported yet')
            if not isinstance(new_loss, torch.Tensor):
                raise NotImplementedError('Multiple losses not supported yet')
            state.loss = self._adjusted_lambda * state.loss + (1 - self._adjusted_lambda) * new_loss


def _gen_indices(x: Tensor) -> Tensor:
    """Generates indices of a random permutation of elements of a batch.

    Args:
        x (torch.Tensor): input tensor of shape ``(B, d1, d2, ..., dn)``,
            B is batch size, d1-dn are feature dimensions.

    Returns:
        indices: A random permutation of the batch indices.
    """
    return torch.randperm(x.shape[0])


def _gen_cutmix_coef(alpha: float) -> float:
    """Generates lambda from ``Beta(alpha, alpha)``.

    Args:
        alpha (float): Parameter for the ``Beta(alpha, alpha)`` distribution.

    Returns:
        cutmix_lambda: Lambda parameter for performing CutMix.
    """
    # First check if alpha is positive.
    assert alpha >= 0
    # Draw the area parameter from a beta distribution.
    # Check here is needed because beta distribution requires alpha > 0
    # but alpha = 0 is fine for cutmix.
    if alpha == 0:
        cutmix_lambda = 0
    else:
        cutmix_lambda = np.random.beta(alpha, alpha)
    return cutmix_lambda


def _rand_bbox(W: int,
               H: int,
               cutmix_lambda: float,
               cx: Optional[int] = None,
               cy: Optional[int] = None,
               uniform_sampling: bool = False) -> Tuple[int, int, int, int]:
    """Randomly samples a bounding box with area determined by ``cutmix_lambda``.

    Adapted from original implementation https://github.com/clovaai/CutMix-PyTorch

    Args:
        W (int): Width of the image
        H (int): Height of the image
        cutmix_lambda (float): Lambda param from cutmix, used to set the area of the
            box if ``cut_w`` or ``cut_h`` is not provided.
        cx (int, optional): Optional x coordinate of the center of the box.
        cy (int, optional): Optional y coordinate of the center of the box.
        uniform_sampling (bool, optional): If true, sample the bounding box such that each pixel
            has an equal probability of being mixed. If false, defaults to the
            sampling used in the original paper implementation. Default: ``False``.

    Returns:
        bbx1: Leftmost edge of the bounding box
        bby1: Top edge of the bounding box
        bbx2: Rightmost edge of the bounding box
        bby2: Bottom edge of the bounding box
    """
    cut_ratio = np.sqrt(1.0 - cutmix_lambda)
    cut_w = int(W * cut_ratio)
    cut_h = int(H * cut_ratio)

    # uniform
    if cx is None:
        if uniform_sampling is True:
            cx = np.random.randint(-cut_w // 2, high=W + cut_w // 2)
        else:
            cx = np.random.randint(W)
    if cy is None:
        if uniform_sampling is True:
            cy = np.random.randint(-cut_h // 2, high=H + cut_h // 2)
        else:
            cy = np.random.randint(H)

    bbx1 = np.clip(cx - cut_w // 2, 0, W)
    bby1 = np.clip(cy - cut_h // 2, 0, H)
    bbx2 = np.clip(cx + cut_w // 2, 0, W)
    bby2 = np.clip(cy + cut_h // 2, 0, H)

    return bbx1, bby1, bbx2, bby2


def _adjust_lambda(x: Tensor, bbox: Tuple) -> float:
    """Rescale the cutmix lambda according to the size of the clipped bounding box.

    Args:
        x (torch.Tensor): input tensor of shape ``(B, d1, d2, ..., dn)``, B is batch size, d1-dn
            are feature dimensions.
        bbox (tuple): (x1, y1, x2, y2) coordinates of the boundind box, obeying x2 > x1, y2 > y1.

    Returns:
        adjusted_lambda: Rescaled cutmix_lambda to account for part of the bounding box
            being potentially out of bounds of the input.
    """
    rx, ry, rw, rh = bbox[0], bbox[1], bbox[2], bbox[3]
    adjusted_lambda = 1 - ((rw - rx) * (rh - ry) / (x.size()[-1] * x.size()[-2]))
    return adjusted_lambda



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Central components used by other modules.

Central parts of composer such as :class:`~.engine.Engine`, base class for critical components such as
:class:`~.algorithm.Algorithm` and :class:`~.callback.Callback` and other useful functionality such as
:class:`~.time.Timestamp` are implemented under core.
"""

from composer.core.algorithm import Algorithm
from composer.core.callback import Callback
from composer.core.data_spec import DataSpec, ensure_data_spec
from composer.core.engine import Engine, Trace
from composer.core.evaluator import Evaluator, ensure_evaluator
from composer.core.event import Event
from composer.core.passes import AlgorithmPass
from composer.core.precision import Precision, get_precision_context
from composer.core.serializable import Serializable
from composer.core.state import State
from composer.core.time import Time, Timestamp, TimeUnit, ensure_time
from composer.core.types import JSON, Batch, BreakEpochException, Dataset, MemoryFormat, PyTorchScheduler, TrainerMode

__all__ = [
    'Algorithm',
    'AlgorithmPass',
    'Callback',
    'DataSpec',
    'Dataset',
    'ensure_data_spec',
    'Engine',
    'Trace',
    'Evaluator',
    'Event',
    'Precision',
    'get_precision_context',
    'Serializable',
    'State',
    'Time',
    'Timestamp',
    'TimeUnit',
    'ensure_time',
    'ensure_evaluator',
    'Batch',
    'PyTorchScheduler',
    'JSON',
    'MemoryFormat',
    'TrainerMode',
    'BreakEpochException',
]



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

# Pytest stub for running lint tests and doctests

# Running these checks through pytest allows us to report any errors in Junit format,
# which is posted directly on the PR

import os
import subprocess
import textwrap

import pytest


def check_output(proc: subprocess.CompletedProcess):
    # Check the subprocess output, and raise an exception with the stdout/stderr dump if there was a non-zero exit
    # The `check=True` flag available in `subprocess.run` does not print stdout/stderr
    if proc.returncode == 0:
        return
    error_msg = textwrap.dedent(f"""\
        Command {proc.args} failed with exit code {proc.returncode}.
        ----Begin stdout----
        {proc.stdout}
        ----End stdout------
        ----Begin stderr----
        {proc.stderr}
        ----End stderr------""")

    raise RuntimeError(error_msg)


@pytest.mark.doctest
def test_run_doctests():
    docs_folder = os.path.join(os.path.dirname(__file__), '..', 'docs')
    check_output(subprocess.run(['make', 'clean'], cwd=docs_folder, capture_output=True, text=True))
    # Must build the html first to ensure that doctests in .. autosummary:: generated pages are included
    check_output(subprocess.run(['make', 'html'], cwd=docs_folder, capture_output=True, text=True))
    check_output(subprocess.run(['make', 'doctest'], cwd=docs_folder, capture_output=True, text=True))


@pytest.mark.doctest
def test_docker_build_matrix():
    """Test that the docker build matrix is up to date."""
    docker_folder = os.path.join(os.path.dirname(__file__), '..', 'docker')

    # Capture the existing readme and build matrix contents
    with open(os.path.join(docker_folder, 'README.md'), 'r') as f:
        existing_readme = f.read()

    with open(os.path.join(docker_folder, 'build_matrix.yaml'), 'r') as f:
        existing_build_matrix = f.read()

    # Run the script
    check_output(
        subprocess.run(['python', os.path.join(docker_folder, 'generate_build_matrix.py')],
                       cwd=docker_folder,
                       capture_output=True,
                       text=True))

    # Assert that the files did not change

    with open(os.path.join(docker_folder, 'README.md'), 'r') as f:
        assert existing_readme == f.read()

    with open(os.path.join(docker_folder, 'build_matrix.yaml'), 'r') as f:
        assert existing_build_matrix == f.read()


@pytest.mark.doctest
@pytest.mark.parametrize('example', [1, 2])
def test_release_tests_reflect_readme(example: int):
    """Test that example_1.py and example_2.py in release_tests reflect the README.md."""
    with open(os.path.join(os.path.dirname(__file__), '..', 'README.md'), 'r') as f:
        readme_lines = f.readlines()
    example_code_lines = []
    found_begin = False
    started = False
    for i, line in enumerate(readme_lines):
        if f'begin_example_{example}' in line:
            found_begin = True
            continue
        # Wait until we get the ```python for start of code snippet
        if found_begin and not started:
            if line == '```python\n':
                started = True
        # Reached end of code snippet
        elif started and line == '```\n':
            # Code snippet continues
            if i + 2 < len(readme_lines) and '-->\n' == readme_lines[
                    i + 1] and '<!--pytest-codeblocks:cont-->\n' == readme_lines[i + 2]:
                started = False
            # Code snippet ends
            else:
                break
        # Add line
        elif started:
            example_code_lines.append(line)

    example_file = os.path.join(os.path.dirname(__file__), '..', '.ci', 'release_tests', f'example_{example}.py')
    with open(example_file, 'r') as f:
        assert f.readlines() == example_code_lines



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Miscellaneous Helpers."""

import socket
from contextlib import contextmanager
from typing import Type

import torch
from torch.nn.parallel import DistributedDataParallel

__all__ = [
    'is_model_deepspeed', 'is_model_fsdp', 'is_notebook', 'warning_on_one_line', 'get_free_tcp_port', 'model_eval_mode'
]


def is_model_deepspeed(model: torch.nn.Module) -> bool:
    """Whether ``model`` is an instance of a :class:`~deepspeed.DeepSpeedEngine`."""
    try:
        import deepspeed
    except ImportError:
        return False
    else:
        return isinstance(model, deepspeed.DeepSpeedEngine)


def is_model_ddp(model: torch.nn.Module) -> bool:
    """Whether ``model`` is an instance of a :class:`.DistributedDataParallel`."""
    return isinstance(model, DistributedDataParallel)


def is_model_fsdp(model: torch.nn.Module) -> bool:
    """Whether ``model`` is an instance of a :class:`.FullyShardedDataParallel`."""
    try:
        from torch.distributed.fsdp import FullyShardedDataParallel as FSDP
        is_fsdp = False
        # Check if model is wrapped with FSDP
        for _, obj in model.named_children():
            if isinstance(obj, FSDP):
                is_fsdp = True
        return is_fsdp
    except ImportError:
        return False


def is_notebook():
    """Whether Composer is running in a IPython/Jupyter Notebook."""
    try:
        __IPYTHON__  #type: ignore
        return True
    except NameError:
        return False


def warning_on_one_line(message: str, category: Type[Warning], filename: str, lineno: int, file=None, line=None):
    """Force Python warnings to consolidate into one line."""
    # From https://stackoverflow.com/questions/26430861/make-pythons-warnings-warn-not-mention-itself
    return f'{category.__name__}: {message} (source: {filename}:{lineno})\n'


def get_free_tcp_port() -> int:
    """Get free socket port to use as MASTER_PORT."""
    # from https://www.programcreek.com/python/?CodeExample=get+free+port
    tcp = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    tcp.bind(('', 0))
    _, port = tcp.getsockname()
    tcp.close()
    return port


@contextmanager
def model_eval_mode(model: torch.nn.Module):
    """Set model.eval() for context duration, restoring model status at end."""
    is_training = model.training
    try:
        model.eval()
        yield
    finally:
        model.train(mode=is_training)



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""DeepLabV3 for image segmentation."""
from composer.models.deeplabv3.model import composer_deeplabv3 as composer_deeplabv3

__all__ = ['composer_deeplabv3']



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Early stopping callback."""

from __future__ import annotations

import logging
from typing import Any, Callable, Optional, Union

import torch

from composer.core import Callback, State, Time, TimeUnit
from composer.loggers import Logger

log = logging.getLogger(__name__)

__all__ = ['EarlyStopper']


class EarlyStopper(Callback):
    """Track a metric and halt training if it does not improve within a given interval.

    Example:
    .. doctest::

        >>> from composer import Evaluator, Trainer
        >>> from composer.callbacks.early_stopper import EarlyStopper
        >>> # constructing trainer object with this callback
        >>> early_stopper = EarlyStopper('MulticlassAccuracy', 'my_evaluator', patience=1)
        >>> evaluator = Evaluator(
        ...     dataloader = eval_dataloader,
        ...     label = 'my_evaluator',
        ...     metric_names = ['MulticlassAccuracy']
        ... )
        >>> trainer = Trainer(
        ...     model=model,
        ...     train_dataloader=train_dataloader,
        ...     eval_dataloader=evaluator,
        ...     optimizers=optimizer,
        ...     max_duration="1ep",
        ...     callbacks=[early_stopper],
        ... )

    Args:
        monitor (str): The name of the metric to monitor.
        dataloader_label (str): The label of the dataloader or evaluator associated with the tracked metric.

            If ``monitor`` is in an :class:`.Evaluator`, the ``dataloader_label`` field should be set to the label of the
            :class:`.Evaluator`.

            If monitor is a training metric or an ordinary evaluation metric not in an :class:`.Evaluator`,
            the ``dataloader_label`` should be set to the dataloader label, which defaults to ``'train'`` or
            ``'eval'``, respectively.
        comp (str | (Any, Any) -> Any, optional): A comparison operator to measure change of the monitored metric.
            The comparison operator will be called ``comp(current_value, prev_best)``. For metrics where the optimal value is low
            (error, loss, perplexity), use a less than operator, and for metrics like accuracy where the optimal value
            is higher, use a greater than operator. Defaults to :func:`torch.less` if loss, error, or perplexity are substrings
            of the monitored metric, otherwise defaults to :func:`torch.greater`.
        min_delta (float, optional): An optional float that requires a new value to exceed the best value by at least that amount.
            Default: ``0.0``.
        patience (Time | int | str, optional): The interval of time the monitored metric can not improve without stopping
            training. Default: 1 epoch. If patience is an integer, it is interpreted as the number of epochs.
    """

    def __init__(
        self,
        monitor: str,
        dataloader_label: str,
        comp: Optional[Union[str, Callable[[
            Any,
            Any,
        ], Any]]] = None,
        min_delta: float = 0.0,
        patience: Union[int, str, Time] = 1,
    ):
        self.monitor = monitor
        self.dataloader_label = dataloader_label
        self.min_delta = abs(min_delta)
        if callable(comp):
            self.comp_func = comp
        if isinstance(comp, str):
            if comp.lower() in ('greater', 'gt'):
                self.comp_func = torch.greater
            elif comp.lower() in ('less', 'lt'):
                self.comp_func = torch.less
            else:
                raise ValueError(
                    "Unrecognized comp string. Use the strings 'gt', 'greater', 'lt' or 'less' or a callable comparison operator"
                )
        if comp is None:
            if any(substr in monitor.lower() for substr in ['loss', 'error', 'perplexity']):
                self.comp_func = torch.less
            else:
                self.comp_func = torch.greater

        self.best = None
        self.best_occurred = None

        if isinstance(patience, str):
            self.patience = Time.from_timestring(patience)
        elif isinstance(patience, int):
            self.patience = Time(patience, TimeUnit.EPOCH)
        else:
            self.patience = patience
            if self.patience.unit not in (TimeUnit.EPOCH, TimeUnit.BATCH):
                raise ValueError('If `patience` is an instance of Time, it must have units of EPOCH or BATCH.')

    def _get_monitored_metric(self, state: State):
        if self.dataloader_label == 'train':
            if self.monitor in state.train_metrics:
                return state.train_metrics[self.monitor].compute()
        else:
            if self.monitor in state.eval_metrics[self.dataloader_label]:
                return state.eval_metrics[self.dataloader_label][self.monitor].compute()
        raise ValueError(f"Couldn't find the metric {self.monitor} with the dataloader label {self.dataloader_label}."
                         "Check that the dataloader_label is set to 'eval', 'train' or the evaluator name.")

    def _update_stopper_state(self, state: State):
        metric_val = self._get_monitored_metric(state)

        if not torch.is_tensor(metric_val):
            metric_val = torch.tensor(metric_val)

        if self.best is None:
            self.best = metric_val
            self.best_occurred = state.timestamp
        elif self.comp_func(metric_val, self.best) and torch.abs(metric_val - self.best) > self.min_delta:
            self.best = metric_val
            self.best_occurred = state.timestamp

        assert self.best_occurred is not None
        if self.patience.unit == TimeUnit.EPOCH:
            if state.timestamp.epoch - self.best_occurred.epoch > self.patience:
                state.stop_training()
        elif self.patience.unit == TimeUnit.BATCH:
            if state.timestamp.batch - self.best_occurred.batch > self.patience:
                state.stop_training()
        else:
            raise ValueError(f'The units of `patience` should be EPOCH or BATCH.')

    def eval_end(self, state: State, logger: Logger) -> None:
        if self.dataloader_label == state.dataloader_label:
            # if the monitored metric is an eval metric or in an evaluator
            self._update_stopper_state(state)

    def epoch_end(self, state: State, logger: Logger) -> None:
        if self.dataloader_label == state.dataloader_label:
            # if the monitored metric is not an eval metric, the right logic is run on EPOCH_END
            self._update_stopper_state(state)

    def batch_end(self, state: State, logger: Logger) -> None:
        if self.patience.unit == TimeUnit.BATCH and self.dataloader_label == state.dataloader_label:
            self._update_stopper_state(state)



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Helpers for the `DeepSpeed <https://www.deepspeed.ai>`_ integration with Composer."""

import copy
import warnings
from typing import Any, Dict, cast

import torch
import torch.utils.data

from composer.core import Batch, Precision, State
from composer.utils import dist, map_collection

__all__ = ['_fix_batch_precision_for_deepspeed', '_parse_deepspeed_config']


def _add_batch_config(config: Dict[str, Any], state: State):
    if state.dataloader is None:
        raise ValueError(
            'When using DeepSpeed, the `train_dataloader` must be specified when constructing the Trainer.')

    try:
        batch_size = state.dataloader.batch_size  # type: ignore as we catch the exception
    except AttributeError as e:
        raise RuntimeError('DeepSpeed requires the `state.dataloader` to have a `batch_size` attribute.') from e

    if state.using_device_microbatch_size:
        assert state.device_train_microbatch_size is not None
        if batch_size % state.device_train_microbatch_size != 0:
            # DeepSpeed will throw an error in this configuration.
            raise ValueError('The Mosaic trainer has been configured to use batch size='
                             f'{batch_size}, but this is not divisible by the '
                             f'train device microbatch size={state.device_train_microbatch_size}. '
                             'This is unsupported when using DeepSpeed.')

        train_batch_size = batch_size * dist.get_world_size()
        # Per the check at the start of this function, the following division is always clean.
        grad_accum = batch_size // state.device_train_microbatch_size

        if 'train_batch_size' in config:
            ds_train_batch_size = config['train_batch_size']
            if ds_train_batch_size != train_batch_size:
                raise ValueError(f'Provided DeepSpeed configuration specifies batch size={ds_train_batch_size}, '
                                 f'but the Mosaic trainer has been configured with batch size={train_batch_size}.')
        else:
            config['train_batch_size'] = train_batch_size

        if 'gradient_accumulation_steps' in config:
            ds_grad_accum = config['gradient_accumulation_steps']
            if ds_grad_accum != state.grad_accum:
                raise ValueError((f'Provided DeepSpeed configuration specifies grad accum={ds_grad_accum}, '
                                  f'but the Mosaic trainer has been configured with grad accum={grad_accum}.'))
        else:
            config['gradient_accumulation_steps'] = state.grad_accum

        if 'train_micro_batch_size_per_gpu' in config:
            ds_per_gpu_microbatch_size = config['train_micro_batch_size_per_gpu']
            if ds_per_gpu_microbatch_size != state.device_train_microbatch_size:
                raise ValueError('Provided DeepSpeed configuration specifies per-GPU microbatch size='
                                 f'{ds_per_gpu_microbatch_size}, but the Mosaic trainer has been '
                                 f'configured with per-GPU microbatch size={state.device_train_microbatch_size}.')
        else:
            config['train_micro_batch_size_per_gpu'] = state.device_train_microbatch_size
    else:
        assert state.grad_accum is not None
        if batch_size % state.grad_accum != 0:
            # DeepSpeed will throw an error in this configuration.
            raise ValueError('The Mosaic trainer has been configured to use batch size='
                             f'{batch_size}, but this is not divisible by the '
                             f'grad accum={state.grad_accum}. This is unsupported when using DeepSpeed.')

        train_batch_size = batch_size * dist.get_world_size()
        # Per the check at the start of this function, the following division is always clean.
        per_gpu_microbatch_size = batch_size // state.grad_accum

        if 'train_batch_size' in config:
            ds_train_batch_size = config['train_batch_size']
            if ds_train_batch_size != train_batch_size:
                raise ValueError(f'Provided DeepSpeed configuration specifies batch size={ds_train_batch_size}, '
                                 f'but the Mosaic trainer has been configured with batch size={train_batch_size}.')
        else:
            config['train_batch_size'] = train_batch_size

        if 'train_micro_batch_size_per_gpu' in config:
            ds_per_gpu_microbatch_size = config['train_micro_batch_size_per_gpu']
            if ds_per_gpu_microbatch_size != per_gpu_microbatch_size:
                raise ValueError('Provided DeepSpeed configuration specifies per-GPU microbatch size='
                                 f'{ds_per_gpu_microbatch_size}, but the Mosaic trainer has been '
                                 f'configured with per-GPU microbatch size={per_gpu_microbatch_size}.')
        else:
            config['train_micro_batch_size_per_gpu'] = per_gpu_microbatch_size

        if 'gradient_accumulation_steps' in config:
            ds_grad_accum = config['gradient_accumulation_steps']
            if ds_grad_accum != state.grad_accum:
                raise ValueError((f'Provided DeepSpeed configuration specifies grad accum={ds_grad_accum}, '
                                  f'but the Mosaic trainer has been configured with grad accum={state.grad_accum}.'))
        else:
            config['gradient_accumulation_steps'] = state.grad_accum


def _ensure_no_optim_in_config(config: Dict[str, Any]):
    if 'optimizer' in config:
        raise ValueError(('The DeepSpeed configuration specifies an optimizer, but the Mosaic '
                          'trainer will override this setting.'))

    if 'scheduler' in config:
        raise ValueError(('The DeepSpeed configuration specifies a scheduler, but the Mosaic '
                          'trainer will override this setting.'))


def _add_precision_config(config: Dict[str, Any], state: State):
    precision = state.precision

    # Verify DeepSpeed config is consistent with state.precision if set. DeepSpeed precision config
    # has many different ways to specify approximately the same thing. See https://www.deepspeed.ai/docs/config-json/.
    ds_precision = None
    if 'fp16' in config and 'enabled' in config['fp16'] and config['fp16']['enabled']:
        ds_precision = Precision.AMP_FP16
    elif 'bf16' in config and 'enabled' in config['bf16'] and config['bf16']['enabled']:
        ds_precision = Precision.AMP_BF16
    elif 'amp' in config and 'enabled' in config['amp'] and config['amp']['enabled']:
        ds_precision = Precision.AMP_FP16
    if ds_precision is not None and ds_precision != precision:
        raise ValueError((f'Provided DeepSpeed configuration specifies precision={ds_precision}, '
                          f'but the Mosaic trainer has been configured with precision={precision}.'))

    # Set DeepSpeed config based on state.precision if not set
    if precision == Precision.AMP_FP16 and 'fp16' not in config:
        config['fp16'] = cast(Dict[str, Any], {'enabled': True})
    elif precision == Precision.AMP_BF16 and 'bf16' not in config:
        config['bf16'] = cast(Dict[str, Any], {'enabled': True})


def _parse_deepspeed_config(
    config: Dict[str, Any],
    state: State,
) -> Dict[str, Any]:
    """Parses the provided DeepSpeed config for compatibility with the Mosaic trainer.

    Broadly speaking, this function does three things.

    1. Check for settings that are unsupported, like DeepSpeed optimizers.

    2. Check for inconsistencies between Mosaic trainer config and DeepSpeed config.

    3. Use Mosaic trainer config to fill in some defaults for DeepSpeed config.

    Args:
        config (Dict[str, Any]): The DeepSpeed config to use. Must follow the format specified
            in `DeepSpeed's documentation <https://www.deepspeed.ai/docs/config-json/>`_.
        state (State): The state of the trainer.

    Returns:
        Dict[str, Any]: The DeepSpeed config updated with values from the arguments passed to the
            :class:`.Trainer`.

    Raises:
        ValueError: If any of the values in the DeepSpeed config conflict with arguments passed
            to the trainer.
        RuntimeError: If the batch size of the train dataloader in the provided state is not set.
    """
    new_config = copy.deepcopy(config)
    _add_batch_config(new_config, state)
    _ensure_no_optim_in_config(new_config)
    _add_precision_config(new_config, state)
    if 'zero_allow_untested_optimizer' in new_config and not new_config['zero_allow_untested_optimizer']:
        warnings.warn(('Provided DeepSpeed configuration specifies zero_allow_untested_optimizer=False. '
                       'This causes DeepSpeed to reject certain Mosaic optimizers that are known to '
                       'work well with DeepSpeed.'))

    new_config['zero_allow_untested_optimizer'] = True
    return new_config


def _convert_fp32_tensor_to_fp16(tensor: torch.Tensor):
    if tensor.dtype == torch.float32:
        return tensor.half()
    return tensor


def _convert_fp32_tensor_to_bf16(tensor: torch.Tensor):
    if tensor.dtype == torch.float32:
        return tensor.to(torch.bfloat16)
    return tensor


def _fix_batch_precision_for_deepspeed(batch: Batch, precision: Precision) -> Batch:
    """Ensures that a batch is properly formatted for DeepSpeed precisions, if active.

    .. note:: Just because the precision is set to FP16 doesn't mean the entire batch can
              be FP16 too. For example, integer tensors are common in inputs and outputs of
              various models, and these must not be converted. The assumption here is
              that a tensor should only be converted to FP16 if it was given in FP32.

    Args:
        batch (Batch): The batch of data to adjust the precision for.
        precision (Precision): The precision to use.

    Returns:
        Batch: The batch with it's precision adjusted to the specified precision.
    """
    if precision == Precision.AMP_FP16:
        return map_collection(batch, _convert_fp32_tensor_to_fp16)  # type: ignore
    elif precision == Precision.AMP_BF16:
        return map_collection(batch, _convert_fp32_tensor_to_bf16)  # type: ignore
    return batch



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

import logging
import textwrap

import torch
import torch.nn.utils.parametrize as parametrize
from torch import nn
from torch.fx import symbolic_trace

from composer.core import Algorithm, Event, State
from composer.loggers import Logger
from composer.utils import module_surgery

log = logging.getLogger(__name__)

__all__ = ['apply_weight_standardization', 'WeightStandardization']


def _standardize_weights(W: torch.Tensor):
    """Function to standardize the input weight ``W``"""
    reduce_dims = list(range(1, W.dim()))
    W_var, W_mean = torch.var_mean(W, dim=reduce_dims, keepdim=True, unbiased=False)
    return (W - W_mean) / (torch.sqrt(W_var + 1e-10))


class WeightStandardizer(nn.Module):
    """Class used to apply weight standardization with torch's parametrization package."""

    def forward(self, W):
        return _standardize_weights(W)


def apply_weight_standardization(module: torch.nn.Module, n_last_layers_ignore: int = 0):
    """`Weight Standardization <https://arxiv.org/abs/1903.10520>`_ standardizes convolutional weights in a model.

    Args:
        module (torch.nn.Module): the torch module whose convolutional weights will be parametrized.
        n_last_layers_ignore (int, optional): the number of layers at the end of the module to not apply weight standardization.
            Default: ``0``.
    """
    modules_to_parametrize = (nn.Conv1d, nn.Conv2d, nn.Conv3d)
    # Attempt to symbolically trace a module, so the results of .modules() will be in the order of execution
    try:
        module_trace = symbolic_trace(module)
    except:
        if n_last_layers_ignore > 0:
            log.warning(
                textwrap.dedent(f"""\
                Module could not be symbolically traced likely due to logic in forward() which is not traceable. Modules
                ignored due to n_last_layers={n_last_layers_ignore} may not actually be the last layers of the network.
                To determine the error, try torch.fx.symbolic_trace(module)."""))
        module_trace = module

    # Count the number of convolution modules in the model
    conv_count = module_surgery.count_module_instances(module_trace, modules_to_parametrize)

    # Calculate how many convs to parametrize based on conv_count and n_last_layers_ignore
    target_ws_count = max(conv_count - n_last_layers_ignore, 0)

    # Parametrize conv modules to use weight standardization
    current_ws_count = 0
    for m in module_trace.modules():
        # If the target number of weight standardized layers is reached, end for loop
        if current_ws_count == target_ws_count:
            break

        if isinstance(m, modules_to_parametrize):
            parametrize.register_parametrization(m, 'weight', WeightStandardizer())
            current_ws_count += 1

    return current_ws_count


class WeightStandardization(Algorithm):
    """`Weight Standardization <https://arxiv.org/abs/1903.10520>`_ standardizes convolutional weights in a model.

    Args:
        n_last_layers_ignore (int, optional): the number of layers at the end of the model to not apply weight standardization.
            Default: ``0``.
    """

    def __init__(self, n_last_layers_ignore: int = 0):
        self.n_last_layers_ignore = n_last_layers_ignore

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(n_last_layers_ignore={self.n_last_layers_ignore})'

    def match(self, event: Event, state: State):
        return (event == Event.INIT)

    def apply(self, event: Event, state: State, logger: Logger):
        count = apply_weight_standardization(state.model, n_last_layers_ignore=self.n_last_layers_ignore)
        logger.log_hyperparameters({'WeightStandardization/num_weights_standardized': count})



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Inference-related utility functions for model export and optimizations.

Used for exporting models into various formats such ONNX, torchscript etc. and apply optimizations such as fusion.
"""
from __future__ import annotations

import contextlib
import copy
import functools
import logging
import os
import tempfile
from typing import TYPE_CHECKING, Any, Callable, Optional, Sequence, Tuple, Union

import torch
import torch.nn as nn

from composer.utils import dist
from composer.utils.checkpoint import download_checkpoint, safe_torch_load
from composer.utils.device import get_device
from composer.utils.iter_helpers import ensure_tuple
from composer.utils.misc import is_model_ddp, is_model_deepspeed, is_model_fsdp, model_eval_mode
from composer.utils.object_store import ObjectStore
from composer.utils.string_enum import StringEnum

if TYPE_CHECKING:
    from composer.devices import Device
    from composer.loggers import Logger

log = logging.getLogger(__name__)

__all__ = ['export_for_inference', 'ExportFormat', 'export_with_logger', 'quantize_dynamic']

Transform = Callable[[nn.Module], nn.Module]

# This is the most common way to use dynamic quantization.
#  Example:
#    from composer.utils import quantize_dynamic
#    export_for_inference(
#        ...
#        transforms = [quantize_dynamic],
#        ...
#    )
#  A user can always redefine it with extra options. This also serves as an example of what to pass to transforms.
quantize_dynamic = functools.partial(torch.quantization.quantize_dynamic, qconfig_spec={torch.nn.Linear})


class ExportFormat(StringEnum):
    """Enum class for the supported export formats.

    Attributes:
        torchscript: Export in "torchscript" format.
        onnx:  Export in "onnx" format.
    """
    TORCHSCRIPT = 'torchscript'
    ONNX = 'onnx'


def _move_sample_input_to_device(sample_input: Optional[Union[torch.Tensor, dict, list, Tuple]],
                                 device: Device) -> Optional[Union[torch.Tensor, dict, list, Tuple]]:
    """Handle moving sample_input of various types to a device. If possible, avoids creating copies of the input."""
    output = None
    if isinstance(sample_input, torch.Tensor):
        output = device.tensor_to_device(sample_input)
    elif isinstance(sample_input, dict):
        for key, value in sample_input.items():
            sample_input[key] = _move_sample_input_to_device(value, device)
        output = sample_input
    elif isinstance(sample_input, list):
        for i in range(len(sample_input)):
            sample_input[i] = _move_sample_input_to_device(sample_input[i], device)
        output = sample_input
    elif isinstance(sample_input, tuple):
        new_tuple = []
        for tuple_item in sample_input:
            new_tuple.append(_move_sample_input_to_device(tuple_item, device))
        output = tuple(new_tuple)

    return output


def export_for_inference(
    model: nn.Module,
    save_format: Union[str, ExportFormat],
    save_path: str,
    save_object_store: Optional[ObjectStore] = None,
    sample_input: Optional[Any] = None,
    dynamic_axes: Optional[Any] = None,
    surgery_algs: Optional[Union[Callable[[nn.Module], nn.Module], Sequence[Callable[[nn.Module], nn.Module]]]] = None,
    transforms: Optional[Sequence[Transform]] = None,
    load_path: Optional[str] = None,
    load_object_store: Optional[ObjectStore] = None,
    load_strict: bool = False,
) -> None:
    """Export a model for inference.

    Args:
        model (nn.Module): An instance of nn.Module. Please note that model is not modified inplace.
            Instead, export-related transformations are applied to a  copy of the model.
        save_format (Union[str, ExportFormat]):  Format to export to. Either ``"torchscript"`` or ``"onnx"``.
        save_path: (str): The path for storing the exported model. It can be a path to a file on the local disk,
        a URL, or if ``save_object_store`` is set, the object name
            in a cloud bucket. For example, ``my_run/exported_model``.
        save_object_store (ObjectStore, optional): If the ``save_path`` is in an object name in a cloud bucket
            (i.e. AWS S3 or Google Cloud Storage), an instance of
            :class:`~.ObjectStore` which will be used
            to store the exported model. Set this to ``None`` if ``save_path`` is a local filepath.
            (default: ``None``)
        sample_input (Any, optional): Example model inputs used for tracing. This is needed for "onnx" export.
            The ``sample_input`` need not match the batch size you intend to use for inference. However, the model
            should accept the ``sample_input`` as is. (default: ``None``)
        dynamic_axes (Any, optional): Dictionary specifying the axes of input/output tensors as dynamic. May be required
            for exporting models using older versions of PyTorch when types cannot be inferred.
        surgery_algs (Union[Callable, Sequence[Callable]], optional): Algorithms that should be applied to the model
            before loading a checkpoint. Each should be callable that takes a model and returns None.
            ``surgery_algs`` are applied before ``transforms``. (default: ``None``)
        transforms (Sequence[Transform], optional): transformations (usually optimizations) that should
            be applied to the model. Each Transform should be a callable that takes a model and returns a modified model.
            ``transforms`` are applied after ``surgery_algs``. (default: ``None``)
        load_path (str): The path to an existing checkpoint file.
            It can be a path to a file on the local disk, a URL, or if ``load_object_store`` is set, the object name
            for a checkpoint in a cloud bucket. For example, run_name/checkpoints/ep0-ba4-rank0. (default: ``None``)
        load_object_store (ObjectStore, optional): If the ``load_path`` is in an object name  in a cloud bucket
            (i.e. AWS S3 or Google Cloud Storage), an instance of
            :class:`~.ObjectStore` which will be used to retreive the checkpoint.
            Otherwise, if the checkpoint is a local filepath, set to ``None``. (default: ``None``)
        load_strict (bool): Whether the keys (i.e., model parameter names) in the model state dict should
            perfectly match the keys in the model instance. (default: ``False``)

    Returns:
        None
    """
    save_format = ExportFormat(save_format)

    if is_model_deepspeed(model):
        raise ValueError(f'Exporting for deepspeed models is currently not supported.')

    if is_model_ddp(model):
        raise ValueError(
            f'Directly exporting a DistributedDataParallel model is not supported. Export the module instead.')

    if is_model_fsdp(model):
        raise ValueError(
            'Directly exporting a FSDP wrapped module is not supported as the model is deepcopied to avoid '
            'side-effects, and FSDP does not support deepcopying. To export the model, load it without FSDP '
            'wrapping.')

    # Only rank0 exports the model
    if dist.get_global_rank() != 0:
        return

    # Make a copy of the model so that we don't modify the original model
    model = copy.deepcopy(model)

    # Make a copy of the sample input so that we don't modify the original sample input
    sample_input = copy.deepcopy(sample_input)

    # Move model and sample input to CPU for export
    cpu = get_device('cpu')
    cpu.module_to_device(model)

    if sample_input is not None:
        sample_input = ensure_tuple(sample_input)
        sample_input = _move_sample_input_to_device(sample_input, cpu)

    # Apply surgery algorithms in the given order
    for alg in ensure_tuple(surgery_algs):
        alg(model)

    if load_path is not None:
        # download checkpoint and load weights only
        log.debug('Loading checkpoint at %s', load_path)
        with tempfile.TemporaryDirectory() as tempdir:
            composer_states_filepath, _, _ = download_checkpoint(path=load_path,
                                                                 node_checkpoint_folder=tempdir,
                                                                 object_store=load_object_store,
                                                                 progress_bar=True)
            state_dict = safe_torch_load(composer_states_filepath)
            missing_keys, unexpected_keys = model.load_state_dict(state_dict['state']['model'], strict=load_strict)
            if len(missing_keys) > 0:
                log.warning(f"Found these missing keys in the checkpoint: {', '.join(missing_keys)}")
            if len(unexpected_keys) > 0:
                log.warning(f"Found these unexpected keys in the checkpoint: {', '.join(unexpected_keys)}")

    with model_eval_mode(model):
        # Apply transformations (i.e., inference optimizations) in the given order
        for transform in ensure_tuple(transforms):
            model = transform(model)

        is_remote_store = save_object_store is not None
        tempdir_ctx = tempfile.TemporaryDirectory() if is_remote_store else contextlib.nullcontext(None)
        with tempdir_ctx as tempdir:
            if is_remote_store:
                local_save_path = os.path.join(str(tempdir), 'model.export')
            else:
                local_save_path = save_path

            if save_format == ExportFormat.TORCHSCRIPT:
                export_model = None
                try:
                    export_model = torch.jit.script(model)
                except Exception:
                    if sample_input is not None:
                        log.warning('Scripting with torch.jit.script failed. Trying torch.jit.trace!',)
                        export_model = torch.jit.trace(model, sample_input)
                    else:
                        log.warning(
                            'Scripting with torch.jit.script failed and sample inputs are not provided for tracing '
                            'with torch.jit.trace',
                            exc_info=True)

                if export_model is not None:
                    torch.jit.save(export_model, local_save_path)
                else:
                    raise RuntimeError('Scritping and tracing failed! No model is getting exported.')

            if save_format == ExportFormat.ONNX:
                if sample_input is None:
                    raise ValueError(f'sample_input argument is required for onnx export')

                input_names = []

                # assert statement for pyright error: Cannot access member "keys" for type "Tensor"
                assert isinstance(sample_input, tuple)
                # Extract input names from sample_input if it contains dicts
                for i in range(len(sample_input)):
                    if isinstance(sample_input[i], dict):
                        input_names += list(sample_input[i].keys())

                # Default input name if no dict present
                if input_names == []:
                    input_names = ['input']

                torch.onnx.export(
                    model,
                    sample_input,
                    local_save_path,
                    input_names=input_names,
                    output_names=['output'],
                    dynamic_axes=dynamic_axes,
                    opset_version=13,
                )

            # upload if required.
            if is_remote_store:
                save_object_store.upload_object(save_path, local_save_path)


def export_with_logger(
    model: nn.Module,
    save_format: Union[str, ExportFormat],
    save_path: str,
    logger: Logger,
    save_object_store: Optional[ObjectStore] = None,
    sample_input: Optional[Any] = None,
    transforms: Optional[Sequence[Transform]] = None,
) -> None:
    """Helper method for exporting a model for inference.

    Exports the model to:
    1) save_object_store, if one is provided,
    2) logger.upload_file(save_path), if (1) does not apply and the logger has a destination that supports file uploading,
    3) locally, if (1) and (2) do not apply.

    Args:
        model (nn.Module): An instance of nn.Module. Please note that model is not modified inplace.
            Instead, export-related transformations are applied to a  copy of the model.
        save_format (Union[str, ExportFormat]):  Format to export to. Either ``"torchscript"`` or ``"onnx"``.
        save_path: (str): The path for storing the exported model. It can be a path to a file on the local disk,
        a URL, or if ``save_object_store`` is set, the object name
            in a cloud bucket. For example, ``my_run/exported_model``.
        logger (Logger): If this logger has a destination that supports file uploading, and save_object_store
            is not provided, this logger is used to export the model.
        save_object_store (ObjectStore, optional): If the ``save_path`` is in an object name in a cloud bucket
            (i.e. AWS S3 or Google Cloud Storage), an instance of
            :class:`~.ObjectStore` which will be used
            to store the exported model. Set this to ``None`` if the logger should be used to export the model or
            if ``save_path`` is a local filepath.
            (default: ``None``)
        sample_input (Any, optional): Example model inputs used for tracing. This is needed for "onnx" export.
            The ``sample_input`` need not match the batch size you intend to use for inference. However, the model
            should accept the ``sample_input`` as is. (default: ``None``)
        transforms (Sequence[Transform], optional): transformations (usually optimizations) that should
            be applied to the model. Each Transform should be a callable that takes a model and returns a modified model.
            ``transforms`` are applied after ``surgery_algs``. (default: ``None``)

    Returns:
        None
    """
    if save_object_store == None and logger.has_file_upload_destination():
        with tempfile.TemporaryDirectory() as tmpdir:
            temp_local_save_path = os.path.join(str(tmpdir), f'model')
            export_for_inference(model=model,
                                 save_format=save_format,
                                 save_path=temp_local_save_path,
                                 sample_input=sample_input,
                                 transforms=transforms)
            logger.upload_file(remote_file_name=save_path, file_path=temp_local_save_path)
    else:
        export_for_inference(model=model,
                             save_format=save_format,
                             save_path=save_path,
                             save_object_store=save_object_store,
                             sample_input=sample_input,
                             transforms=transforms)



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Implements ViT-S/16 as a :class:`.ComposerClassifier`."""

from composer.models.tasks import ComposerClassifier

__all__ = ['vit_small_patch16']


def vit_small_patch16(num_classes: int = 1000,
                      image_size: int = 224,
                      channels: int = 3,
                      dropout: float = 0.0,
                      embedding_dropout: float = 0.0):
    """Helper function to create a :class:`.ComposerClassifier` using a ViT-S/16 model.

    See `Training data-efficient image transformers & distillation through attention <https://arxiv.org/pdf/2012.12877.pdf>`_
        (Touvron et al, 2021) for details on ViT-S/16.

    Args:
        num_classes (int, optional): number of classes for the model. Default: ``1000``.
        image_size (int, optional): input image size. If you have rectangular images, make sure your image
         size is the maximum of the width and height. Default: ``224``.
        channels (int, optional): number of  image channels. Default: ``3``.
        dropout (float, optional): 0.0 - 1.0 dropout rate. Default: ``0``.
        embedding_dropout (float, optional): 0.0 - 1.0 embedding dropout rate. Default: ``0``.

    Returns:
        ComposerModel: instance of :class:`.ComposerClassifier` with a ViT-S/16 model.
    """

    from vit_pytorch import ViT
    model = ViT(
        image_size=image_size,
        channels=channels,
        num_classes=num_classes,
        dim=384,  # embed dim/width
        patch_size=16,
        depth=12,  # layers
        heads=6,
        mlp_dim=1536,
        dropout=dropout,
        emb_dropout=embedding_dropout)

    composer_model = ComposerClassifier(module=model)
    return composer_model



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Logs metrics to dictionary objects that persist in memory throughout training.

Useful for collecting and plotting data inside notebooks.
"""

from __future__ import annotations

import copy
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple

import numpy as np
from torch import Tensor

from composer.core.time import Time
from composer.loggers.logger import Logger
from composer.loggers.logger_destination import LoggerDestination

if TYPE_CHECKING:
    from composer.core import State, Timestamp

__all__ = ['InMemoryLogger']


class InMemoryLogger(LoggerDestination):
    """Logs metrics to dictionary objects that persist in memory throughout training.

    Useful for collecting and plotting data inside notebooks.

    Example usage:
        .. testcode::

            from composer.loggers import InMemoryLogger
            from composer.trainer import Trainer
            logger = InMemoryLogger(
            )
            trainer = Trainer(
                model=model,
                train_dataloader=train_dataloader,
                eval_dataloader=eval_dataloader,
                max_duration="1ep",
                optimizers=[optimizer],
                loggers=[logger]
            )
            # Get data from logger. If you are using multiple loggers, be sure to confirm
            # which index in trainer.logger.destinations contains your desired logger.
            logged_data = trainer.logger.destinations[0].data

    Attributes:
        data (Dict[str, List[Tuple[Timestamp, Any]]]): Mapping of a logged key to
            a (:class:`~.time.Timestamp`, logged value) tuple.
            This dictionary contains all logged data.
        most_recent_values (Dict[str, Any]): Mapping of a key to the most recent value for that key.
        most_recent_timestamps (Dict[str, Timestamp]): Mapping of a key to the
            :class:`~.time.Timestamp` of the last logging call for that key.
        hyperparameters (Dict[str, Any]): Dictionary of all hyperparameters.

    """

    def __init__(self) -> None:
        self.data: Dict[str, List[Tuple[Timestamp, Any]]] = {}
        self.most_recent_values = {}
        self.most_recent_timestamps: Dict[str, Timestamp] = {}
        self.state: Optional[State] = None
        self.hyperparameters: Dict[str, Any] = {}

    def log_hyperparameters(self, hyperparameters: Dict[str, Any]):
        self.hyperparameters.update(hyperparameters)

    def log_metrics(self, metrics: Dict[str, Any], step: Optional[int] = None) -> None:
        assert self.state is not None
        timestamp = self.state.timestamp
        copied_metrics = copy.deepcopy(metrics)
        for k, v in copied_metrics.items():
            if k not in self.data:
                self.data[k] = []
            self.data[k].append((timestamp, v))
        self.most_recent_values.update(copied_metrics.items())
        self.most_recent_timestamps.update({k: timestamp for k in copied_metrics})

    def init(self, state: State, logger: Logger) -> None:
        self.state = state

    def get_timeseries(self, metric: str) -> Dict[str, Any]:
        """Returns logged data as dict containing values of a desired metric over time.

        Args:
            metric (str): Metric of interest. Must be present in self.data.keys().

        Returns:
            timeseries (Dict[str, Any]): Dictionary in which one key is ``metric``,
                and the associated value is a list of values of that metric. The remaining
                keys are each a unit of time, and the associated values are each a list of
                values of that time unit for the corresponding index of the metric. For
                example:
                >>> InMemoryLogger.get_timeseries(metric="accuracy/val")
                {"accuracy/val": [31.2, 45.6, 59.3, 64.7, "epoch": [1, 2, 3, 4, ...],
                ...], "batch": [49, 98, 147, 196, ...], ...}

        Example:
            .. testcode::

                import matplotlib.pyplot as plt

                from composer.loggers import InMemoryLogger
                from composer.core.time import Time, Timestamp

                in_mem_logger = InMemoryLogger()
                trainer = Trainer(
                    model=model,
                    train_dataloader=train_dataloader,
                    eval_dataloader=eval_dataloader,
                    max_duration="1ep",
                    optimizers=[optimizer],
                    loggers=[in_mem_logger]
                )

                # Populate the logger with data
                for b in range(0,3):
                    datapoint = b * 3
                    in_mem_logger.log_metrics({"accuracy/val": datapoint})

                timeseries = in_mem_logger.get_timeseries("accuracy/val")
                plt.plot(timeseries["batch"], timeseries["accuracy/val"])
                plt.xlabel("Batch")
                plt.ylabel("Validation Accuracy")
        """
        # Check that desired metric is in present data
        if metric not in self.data.keys():
            raise ValueError(f'Invalid value for argument `metric`: {metric}. Requested '
                             'metric is not present in self.data.keys().')

        timeseries = {}
        # Iterate through datapoints
        for datapoint in self.data[metric]:
            timestamp, metric_value = datapoint
            timeseries.setdefault(metric, []).append(metric_value)
            # Iterate through time units and add them all!
            for field, time in timestamp.get_state().items():
                time_value = time.value if isinstance(time, Time) else time.total_seconds()
                timeseries.setdefault(field, []).append(time_value)
        # Convert to numpy arrays
        for k, v in timeseries.items():
            if isinstance(v[0], Tensor):
                v = Tensor(v).numpy()
            else:
                v = np.array(v)
            timeseries[k] = v
        return timeseries



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Helpers for running distributed data parallel training."""

import collections
import logging
import warnings
from contextlib import contextmanager, nullcontext
from typing import Any, Callable, ContextManager, Dict, Optional, Sequence, Union, cast

import torch
from packaging import version
from torch.nn.parallel import DistributedDataParallel
from torchmetrics import Metric, MetricCollection

from composer.core import Precision
from composer.core.state import State
from composer.trainer.meta_safe_apply import meta_safe_apply
from composer.utils import StringEnum, dist, ensure_tuple

__all__ = ['DDPSyncStrategy', 'ddp_sync_context', 'prepare_ddp_module', 'prepare_fsdp_module']

log = logging.getLogger(__name__)


class DDPSyncStrategy(StringEnum):
    """How and when gradient synchronization should happen.

    Attributes:
        SINGLE_AUTO_SYNC: The default behavior. Gradients are synchronized as they
            computed, for only the final microbatch of a batch. This is the most efficient
            strategy, but can lead to errors when ``find_unused_parameters`` is set, since
            it is possible different microbatches may use different sets of parameters,
            leading to an incomplete sync.
        MULTI_AUTO_SYNC: The default behavior when ``find_unused_parameters`` is set.
            Gradients are synchronized as they are computed for all microbatches. This ensures
            complete synchronization, but is less efficient than :attr:`SINGLE_AUTO_SYNC`. This
            efficiency gap is usually small, as long as either DDP syncs are a small portion
            of the trainer's overall runtime, or the number of microbatches per batch is
            relatively small.
        FORCED_SYNC: Gradients are manually synchronized only after all gradients have been
            computed for the final microbatch of a batch. Like :attr:`MULTI_AUTO_SYNC`, this
            strategy ensures complete gradient synchronization, but this tends to be slower than
            :attr:`MULTI_AUTO_SYNC`. This is because ordinarily syncs can happen in parallel
            with the ``loss.backward()`` computation, meaning syncs can be mostly complete by
            the time that function finishes. However, in certain circumstances, syncs may take
            a very long time to complete - if there are also a lot of microbatches per batch,
            this strategy may be optimal.
    """
    SINGLE_AUTO_SYNC = 'single_auto_sync'
    MULTI_AUTO_SYNC = 'multi_auto_sync'
    FORCED_SYNC = 'forced_sync'


@contextmanager
def ddp_sync_context(state: State, is_final_microbatch: bool, sync_strategy: Union[str, DDPSyncStrategy]):
    """A context manager for handling the :class:`DDPSyncStrategy`.

    Args:
        state (State): The state of the :class:`.Trainer`.
        is_final_microbatch (bool): Whether or not the context is being used during the final
            microbatch of the gradient accumulation steps.
        sync_strategy (str | DDPSyncStrategy): The ddp sync strategy to use. If a string
            is provided, the string must be one of the values in :class:`DDPSyncStrategy`.
    """
    if not isinstance(state.model, DistributedDataParallel):
        yield
        return

    assert state.optimizers is not None, 'optimizers have not been initialized'
    sync_strategy = DDPSyncStrategy(sync_strategy)

    no_sync_context = cast(Callable[[], ContextManager], state.model.no_sync)
    auto_sync_context = nullcontext

    if sync_strategy == DDPSyncStrategy.SINGLE_AUTO_SYNC:
        context = auto_sync_context if is_final_microbatch else no_sync_context
        with context():
            yield

    elif sync_strategy == DDPSyncStrategy.MULTI_AUTO_SYNC:
        with auto_sync_context():
            yield

    elif sync_strategy == DDPSyncStrategy.FORCED_SYNC:
        try:
            with no_sync_context():
                yield
        finally:
            if is_final_microbatch:
                for optimizer in state.optimizers:
                    for group in optimizer.param_groups:
                        for p in group['params']:
                            if p.grad is not None:
                                dist.all_reduce(p.grad)
                                p.grad = p.grad / dist.get_world_size()

    else:
        raise ValueError('Unknown sync strategy', sync_strategy)


def prepare_ddp_module(module: torch.nn.Module, find_unused_parameters: bool) -> torch.nn.Module:
    """Wraps the module in a :class:`torch.nn.parallel.DistributedDataParallel` object if running distributed training.

    Args:
        module (torch.nn.Module): The module to wrap.
        find_unused_parameters (bool): Whether or not to do a pass over the autograd graph
            to find parameters to not expect gradients for. This is useful if there are some
            parameters in the model that are not being trained.
    """
    if dist.is_available() and dist.is_initialized():
        if any((p.requires_grad for p in module.parameters())):
            log.debug('Wrapping model with DistributedDataParallel')
            ddp_model = DistributedDataParallel(module, find_unused_parameters=find_unused_parameters)
            return ddp_model
        return module
    if dist.is_available():
        raise RuntimeError('Please call dist.initialize_dist() before calling ddp.prepare_module()')

    raise RuntimeError('When the world size is > 1, ``torch.distributed`` must be used. However, it is '
                       'not available in your installation of PyTorch. Please install or build PyTorch '
                       'with distributed support.')


def prepare_fsdp_module(model: torch.nn.Module, optimizers: Optional[Union[torch.optim.Optimizer,
                                                                           Sequence[torch.optim.Optimizer]]],
                        fsdp_config: Dict[str, Any], precision: Precision) -> None:
    """Prepare a module (assumed ComposerModel) and optimizer for use with :class:`torch.distributed.fsdp.FullyShardedDataParallel`.

    Args:
        model (torch.nn.Module): The model to wrap.
        optimizers (torch.optim.Optimizer | Sequence[torch.optim.Optimizer], optional): The optimizer for `model`, assumed to have a single param group := model.parameters().
        fsdp_config (Dict[str, Any]): The FSDP config.
        precision: (Precision): The precision being used by the Trainer, used to fill in defaults for FSDP `mixed_precision` settings.
    """
    if version.parse(torch.__version__) < version.parse('1.13.0'):
        raise RuntimeError('To use FSDP with Composer, you must use torch>=1.13.0.')
    from torch.distributed.algorithms._checkpoint.checkpoint_wrapper import (CheckpointImpl,
                                                                             apply_activation_checkpointing,
                                                                             checkpoint_wrapper)
    from torch.distributed.fsdp import FullyShardedDataParallel
    from torch.distributed.fsdp.flatten_params_wrapper import FlattenParamsWrapper

    from composer.trainer.mosaic_fsdp import (MosaicFullyShardedDataParallel, backward_prefetch_map, get_cpu_offload,
                                              get_mixed_precision, sharding_map)

    if optimizers:
        optimizers_tuple = ensure_tuple(optimizers)
        if len(optimizers_tuple) != 1:
            raise NotImplementedError(f'Only one optimizer is supported; found {len(optimizers_tuple)} optimizers')

        # clearing optimizer param groups and state
        # that will be recreated at the end of prepare_fsdp_module
        optim = optimizers_tuple[0]
        optim.param_groups.clear()
        optim.state.clear()

    sharding_map_key = fsdp_config.get('sharding_strategy', 'FULL_SHARD').upper()
    sharding_strategy = sharding_map[sharding_map_key]

    cpu_offload = get_cpu_offload(cpu_offload=fsdp_config.get('cpu_offload', False))

    mixed_precision = fsdp_config.get('mixed_precision', 'DEFAULT')
    keep_low_precision_grads = fsdp_config.get('keep_low_precision_grads', False)
    mixed_precision, param_dtype, _, _ = get_mixed_precision(precision,
                                                             mixed_precision=mixed_precision,
                                                             keep_low_precision_grads=keep_low_precision_grads)

    # Note: FSDP does support the use of torch.float32 with sharding.
    # They just never expected a user to pass in torch.float32 into mixed_precision as a param_dtype.
    # See: https://github.com/pytorch/pytorch/issues/90584
    # The PR fixing this bug is merged into PyTorch, but it hasn't made its way into a release yet.
    # Instead a user needs to pass in `None` as param_dtype to have the parameters as torch.float32.
    # TODO: remove these checks when PyTorch has a release that includes the fix.
    if sharding_map_key != 'NO_SHARD':
        if (precision == Precision.AMP_FP16 and param_dtype not in [torch.float16, None] or
                precision == Precision.AMP_BF16 and param_dtype not in [torch.bfloat16, None]):
            raise ValueError(
                f'FSDP in PyTorch 1.13 does not support precision `{precision}` with sharding strategy `{sharding_strategy}` '
                f'and param_dtype `{param_dtype}.` Consider using one of the predefined mixed_precision strategies '
                "(choose: `'FULL'`, `'DEFAULT'`, `'PURE'`)")

        if param_dtype == torch.float32:
            raise ValueError(
                f'FSDP in PyTorch 1.13 does not support param_dtype `{param_dtype}` with sharding_strategy `{sharding_map_key}` '
                f'Consider using `amp` or `bf16` for precision or setting param_dtype in mixed_precision to `None` '
                f'with sharding strategy `{sharding_map_key}.`')

    backward_prefetch = backward_prefetch_map[fsdp_config.get('backward_prefetch', 'BACKWARD_POST').upper()]
    min_params = int(float(fsdp_config.get('min_params', 1e9)))
    activation_checkpointing = fsdp_config.get('activation_checkpointing', False)
    activation_cpu_offload = fsdp_config.get('activation_cpu_offload', False)
    sync_module_states = fsdp_config.get('sync_module_states', False)
    forward_prefetch = fsdp_config.get('forward_prefetch', False)
    limit_all_gathers = fsdp_config.get('limit_all_gathers', False)
    ignored_modules = fsdp_config.get('ignored_modules', None)
    state_dict_type = fsdp_config.get('state_dict_type', 'full')
    activation_checkpointing_reentrant = fsdp_config.get('activation_checkpointing_reentrant', True)

    # We choose to not wrap the ComposerModel directly, but instead wrap any submodules like `ComposerModel.model`
    # This makes it safer to call ComposerModel-specific functions like 'eval_forward' that
    # may make calls to sharded submodules. If we only wrap the submodules, then any call that ComposerModel makes
    # to a FSDP-wrapped submodule's `forward()` function will be safe and all-gather the necessary weights before `forward()`.
    for obj_name, obj in model.named_children():
        if not isinstance(obj, (Metric, MetricCollection)):

            def _param_init_fn(module: torch.nn.Module) -> None:
                # A dictionary of all tied parameter pointers to module names
                tied_pointers = {}

                # Goes through all modules finding which weights have the same pointers
                for name, mod in module.named_modules():
                    for attr in ['weight', 'bias']:
                        if hasattr(mod, attr):
                            ptr = id(getattr(mod, attr))
                            ptr_attr = (ptr, attr)
                            name_list = tied_pointers.get(ptr_attr, [])
                            name_list.append(name)
                            tied_pointers[ptr_attr] = name_list

                # Creates a dictionary of module names that should be tied together
                tied_mod_names = collections.defaultdict(list)
                # Creates a set of modules we should not initialize
                should_not_init_params = set()
                for ptr_attr_type, mod_names in tied_pointers.items():
                    # No modules for this pointer are tied
                    if len(mod_names) == 1:
                        continue
                    _, attr_type = ptr_attr_type
                    first = next(mod_names.__iter__())
                    for elem in mod_names:
                        should_not_init_params.add('.'.join([elem, attr_type]))
                        tied_mod_names[(first, attr_type)].append(elem)
                    # Make sure at least one of the tied parameters is initialized
                    should_not_init_params.remove('.'.join([first, attr_type]))

                meta_safe_apply(module,
                                lambda t: torch.empty_like(t, device=f'cuda:{torch.cuda.current_device()}'),
                                should_not_init_params,
                                module_name='')

                if len(tied_mod_names) > 0:
                    warnings.warn(('The passed in model appears to have tied weights. In order to '
                                   'support effective weight tying, the tied modules need to be '
                                   'in the same FSDP module. If the weights are not properly tied '
                                   'it can lead to loss spikes. We have tried our best to ensure '
                                   'the tied weights are in the same FSDP module.'))

                # Redoes weight tying
                for name_attr, tied_names in tied_mod_names.items():
                    name, attr = name_attr
                    src_mod = module.get_submodule(name)
                    # We need to make sure the source and destination
                    # modules end up in the same FSDP module otherwise
                    # with sharding weight tying gets violated
                    src_mod._fsdp_wrap = False  # type: ignore
                    src_params = getattr(src_mod, attr)
                    for tied_name in tied_names:
                        dest_mod = module.get_submodule(tied_name)
                        dest_mod._fsdp_wrap = False  # type: ignore
                        setattr(dest_mod, attr, src_params)

                if hasattr(obj, 'param_init_fn') and isinstance(obj.param_init_fn, Callable):
                    module.apply(obj.param_init_fn)
                elif hasattr(module, 'reset_parameters') and isinstance(module.reset_parameters, Callable):
                    module.reset_parameters()
                else:
                    raise ValueError(
                        f'Object `{obj_name}` does not have a ``param_init_fn`` or a ``reset_parameters`` function. '
                        'This leaves parameters without initialization. Please add a ``param_init_fn`` or ``reset_parameters`` '
                        f'to module `{obj_name}`.')

            # Choose which modules to FSDP wrap according to the following priority:
            # If module has attribute `module._fsdp_wrap = ...`, always respect it
            # Otherwise wrap if root object `obj.fsdp_wrap_fn(module)` is true
            # Or if unwrapped params in module in greater than or equal to fsdp_config.min_params
            def _auto_wrap_policy(module: torch.nn.Module, recurse: bool, unwrapped_params: int) -> bool:
                if recurse:
                    return True
                else:
                    if hasattr(module, '_fsdp_wrap'):
                        return bool(module._fsdp_wrap)

                    is_large = unwrapped_params >= min_params
                    if hasattr(obj, 'fsdp_wrap_fn') and isinstance(obj.fsdp_wrap_fn, Callable):
                        return obj.fsdp_wrap_fn(module) or is_large
                    else:
                        return is_large

            fsdp_obj = MosaicFullyShardedDataParallel(
                obj,
                sharding_strategy=sharding_strategy,
                auto_wrap_policy=_auto_wrap_policy,
                cpu_offload=cpu_offload,
                mixed_precision=mixed_precision,
                backward_prefetch=backward_prefetch,
                ignored_modules=ignored_modules,
                param_init_fn=_param_init_fn,
                device_id=torch.cuda.current_device(),
                sync_module_states=sync_module_states,
                forward_prefetch=forward_prefetch,
                limit_all_gathers=limit_all_gathers,
            )

            # Activation Checkpointing
            if activation_checkpointing or activation_cpu_offload:
                if not activation_checkpointing_reentrant:
                    first_wrap_fn = lambda m: checkpoint_wrapper(m, checkpoint_impl=CheckpointImpl.NO_REENTRANT
                                                                ) if activation_checkpointing else (lambda module:
                                                                                                    module)
                    second_wrap_fn = (
                        lambda module: checkpoint_wrapper(
                            first_wrap_fn(module),  # type: ignore reportGeneralTypeIssues
                            checkpoint_impl=CheckpointImpl.NO_REENTRANT,
                            offload_to_cpu=True)) if activation_cpu_offload else first_wrap_fn
                else:
                    first_wrap_fn = checkpoint_wrapper if activation_checkpointing else (lambda module: module)
                    second_wrap_fn = (
                        lambda module: checkpoint_wrapper(
                            first_wrap_fn(module),  # type: ignore reportGeneralTypeIssues
                            offload_to_cpu=True)) if activation_cpu_offload else first_wrap_fn

                # Choose which modules to activation checkpoint according to the following priority:
                # If module has attribute `module._activation_checkpointing = ...`, always respect it
                # Otherwise checkpoint if root object `obj.activation_checkpointing_fn(module)` is true
                def _check_fn(module: torch.nn.Module) -> bool:
                    if isinstance(module, (FullyShardedDataParallel, FlattenParamsWrapper)):
                        return False
                    if hasattr(module, '_activation_checkpointing'):
                        return bool(module._activation_checkpointing)
                    if hasattr(obj, 'activation_checkpointing_fn') and isinstance(obj.activation_checkpointing_fn,
                                                                                  Callable):
                        return obj.activation_checkpointing_fn(module)
                    return False

                apply_activation_checkpointing(
                    fsdp_obj,
                    checkpoint_wrapper_fn=second_wrap_fn,  # type: ignore
                    check_fn=_check_fn,  # type: ignore
                )

            setattr(model, obj_name, fsdp_obj)

    # Print FSDP wrapped model and FSDP config if `verbose=True`
    if fsdp_config.get('verbose', False):
        print(f'FSDP: Wrapped Model:')
        print(model)
        print(f'FSDP: Using sharding_strategy={sharding_strategy}')
        print(f'FSDP: Using cpu_offload={cpu_offload}')
        print(f'FSDP: Using mixed_precision={mixed_precision}')
        print(f'FSDP: Using backward_prefetch={backward_prefetch}')
        print(f'FSDP: Using min_params={min_params}')
        print(f'FSDP: Using activation_checkpointing={activation_checkpointing}')
        print(f'FSDP: Using activation_cpu_offload={activation_cpu_offload}')
        print(f'FSDP: Using sync_module_states={sync_module_states}')
        print(f'FSDP: Using forward_prefetch={forward_prefetch}')
        print(f'FSDP: Using limit_all_gathers={limit_all_gathers}')
        print(f'FSDP: Using state_dict_type={state_dict_type}')

    # Rebuild optimizer now that parameters are sharded
    if optimizers:
        optimizers_tuple = ensure_tuple(optimizers)
        optim = optimizers_tuple[0]
        optim.param_groups.clear()
        optim.add_param_group({'params': list(model.parameters())})



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Training Loop Events."""
from composer.utils import StringEnum

__all__ = ['Event']


class Event(StringEnum):
    """Enum to represent training loop events.

    Events mark specific points in the training loop where an :class:`~.core.Algorithm` and :class:`~.core.Callback`
    can run.

    The following pseudocode shows where each event fires in the training loop:

    .. code-block:: python

        # <INIT>
        # <AFTER_LOAD>
        # <FIT_START>
        for epoch in range(NUM_EPOCHS):
            # <EPOCH_START>
            while True:
                # <BEFORE_DATALOADER>
                batch = next(dataloader)
                if batch is None:
                    break
                # <AFTER_DATALOADER>

                # <BATCH_START>

                # <BEFORE_TRAIN_BATCH>

                for microbatch in batch.split(grad_accum):

                    # <BEFORE_FORWARD>
                    outputs = model(batch)
                    # <AFTER_FORWARD>

                    # <BEFORE_LOSS>
                    loss = model.loss(outputs, batch)
                    # <AFTER_LOSS>

                    # <BEFORE_BACKWARD>
                    loss.backward()
                    # <AFTER_BACKWARD>

                # Un-scale and clip gradients

                # <AFTER_TRAIN_BATCH>
                optimizer.step()

                # <BATCH_END>

                if should_eval(batch=True):
                    for eval_dataloader in eval_dataloaders:
                        # <EVAL_START>
                        for batch in eval_dataloader:
                            # <EVAL_BATCH_START>
                            # <EVAL_BEFORE_FORWARD>
                            outputs, targets = model(batch)
                            # <EVAL_AFTER_FORWARD>
                            metrics.update(outputs, targets)
                            # <EVAL_BATCH_END>
                        # <EVAL_END>

                # <BATCH_CHECKPOINT>
            # <EPOCH_END>

            if should_eval(batch=False):
                for eval_dataloader in eval_dataloaders:
                    # <EVAL_START>
                    for batch in eval_dataloader:
                        # <EVAL_BATCH_START>
                        # <EVAL_BEFORE_FORWARD>
                        outputs, targets = model(batch)
                        # <EVAL_AFTER_FORWARD>
                        metrics.update(outputs, targets)
                        # <EVAL_BATCH_END>
                    # <EVAL_END>

            # <EPOCH_CHECKPOINT>
        # <FIT_END>

    Attributes:
        INIT: Invoked in the constructor of :class:`~.trainer.Trainer`. Model surgery (see
            :mod:`~composer.utils.module_surgery`) typically occurs here.
        AFTER_LOAD: Immediately after checkpoint is loaded in constructor of :class:`~.trainer.Trainer`.
        FIT_START: Invoked at the beginning of each call to :meth:`.Trainer.fit`. Dataset transformations typically
            occur here.
        EPOCH_START: Start of an epoch.
        BEFORE_DATALOADER: Immediately before the dataloader is called.
        AFTER_DATALOADER: Immediately after the dataloader is called.  Typically used for on-GPU dataloader transforms.
        BATCH_START: Start of a batch.
        BEFORE_TRAIN_BATCH: Before the forward-loss-backward computation for a training batch. When using gradient
            accumulation, this is still called only once.
        BEFORE_FORWARD: Before the call to ``model.forward()``.
            This is called multiple times per batch when using gradient accumulation.
        AFTER_FORWARD: After the call to ``model.forward()``.
            This is called multiple times per batch when using gradient accumulation.
        BEFORE_LOSS: Before the call to ``model.loss()``.
            This is called multiple times per batch when using gradient accumulation.
        AFTER_LOSS: After the call to ``model.loss()``.
            This is called multiple times per batch when using gradient accumulation.
        BEFORE_BACKWARD: Before the call to ``loss.backward()``.
            This is called multiple times per batch when using gradient accumulation.
        AFTER_BACKWARD: After the call to ``loss.backward()``.
            This is called multiple times per batch when using gradient accumulation.
        AFTER_TRAIN_BATCH: After the forward-loss-backward computation for a training batch. When using gradient
            accumulation, this event still fires only once.
        BATCH_END: End of a batch, which occurs after the optimizer step and any gradient scaling.
        BATCH_CHECKPOINT: After :attr:`.Event.BATCH_END` and any batch-wise evaluation. Saving checkpoints at this
            event allows the checkpoint saver to use the results from any batch-wise evaluation to determine whether
            a checkpoint should be saved.
        EPOCH_END: End of an epoch.
        EPOCH_CHECKPOINT: After :attr:`.Event.EPOCH_END` and any epoch-wise evaluation. Saving checkpoints at this
            event allows the checkpoint saver to use the results from any epoch-wise evaluation to determine whether
            a checkpointshould be saved.
        FIT_END: Invoked at the end of each call to :meth:`.Trainer.fit`. This event exists primarily for logging information
            and flushing callbacks. Algorithms should not transform the training state on this event, as any changes will not
            be preserved in checkpoints.

        EVAL_START: Start of evaluation through the validation dataset.
        EVAL_BATCH_START: Before the call to ``model.eval_forward(batch)``
        EVAL_BEFORE_FORWARD: Before the call to ``model.eval_forward(batch)``
        EVAL_AFTER_FORWARD: After the call to ``model.eval_forward(batch)``
        EVAL_BATCH_END: After the call to ``model.eval_forward(batch)``
        EVAL_END: End of evaluation through the validation dataset.
    """

    INIT = 'init'
    AFTER_LOAD = 'after_load'
    FIT_START = 'fit_start'

    EPOCH_START = 'epoch_start'

    BEFORE_DATALOADER = 'before_dataloader'
    AFTER_DATALOADER = 'after_dataloader'

    BATCH_START = 'batch_start'

    BEFORE_TRAIN_BATCH = 'before_train_batch'

    BEFORE_FORWARD = 'before_forward'
    AFTER_FORWARD = 'after_forward'

    BEFORE_LOSS = 'before_loss'
    AFTER_LOSS = 'after_loss'

    BEFORE_BACKWARD = 'before_backward'
    AFTER_BACKWARD = 'after_backward'

    AFTER_TRAIN_BATCH = 'after_train_batch'

    BATCH_END = 'batch_end'
    BATCH_CHECKPOINT = 'batch_checkpoint'

    EPOCH_END = 'epoch_end'
    EPOCH_CHECKPOINT = 'epoch_checkpoint'

    FIT_END = 'fit_end'

    EVAL_START = 'eval_start'
    EVAL_BATCH_START = 'eval_batch_start'
    EVAL_BEFORE_FORWARD = 'eval_before_forward'
    EVAL_AFTER_FORWARD = 'eval_after_forward'
    EVAL_BATCH_END = 'eval_batch_end'
    EVAL_END = 'eval_end'

    PREDICT_START = 'predict_start'
    PREDICT_BATCH_START = 'predict_batch_start'
    PREDICT_BEFORE_FORWARD = 'predict_before_forward'
    PREDICT_AFTER_FORWARD = 'predict_after_forward'
    PREDICT_BATCH_END = 'predict_batch_end'
    PREDICT_END = 'predict_end'

    @property
    def is_before_event(self) -> bool:
        """Whether the event is an "before" event.

        An "before" event (e.g., :attr:`~Event.BEFORE_LOSS`) has a corresponding "after" event
        (.e.g., :attr:`~Event.AFTER_LOSS`).
        """
        return self in _BEFORE_EVENTS

    @property
    def is_after_event(self) -> bool:
        """Whether the event is an "after" event.

        An "after" event (e.g., :attr:`~Event.AFTER_LOSS`) has a corresponding "before" event
        (.e.g., :attr:`~Event.BEFORE_LOSS`).
        """
        return self in _AFTER_EVENTS

    @property
    def canonical_name(self) -> str:
        """The name of the event, without before/after markers.

        Events that have a corresponding "before" or "after" event share the same canonical name.

        Example:
            >>> Event.EPOCH_START.canonical_name
            'epoch'
            >>> Event.EPOCH_END.canonical_name
            'epoch'

        Returns:
            str: The canonical name of the event.
        """
        name: str = self.value
        name = name.replace('before_', '')
        name = name.replace('after_', '')
        name = name.replace('_start', '')
        name = name.replace('_end', '')
        return name

    @property
    def is_predict(self) -> bool:
        """Whether the event is during the predict loop."""
        return self.value.startswith('predict')

    @property
    def is_eval(self) -> bool:
        """Whether the event is during the eval loop."""
        return self.value.startswith('eval')


_BEFORE_EVENTS = (Event.FIT_START, Event.EPOCH_START, Event.BEFORE_DATALOADER, Event.BATCH_START,
                  Event.BEFORE_TRAIN_BATCH, Event.BEFORE_FORWARD, Event.BEFORE_LOSS, Event.BEFORE_BACKWARD,
                  Event.EVAL_START, Event.EVAL_BATCH_START, Event.EVAL_BEFORE_FORWARD, Event.PREDICT_START,
                  Event.PREDICT_BATCH_START, Event.PREDICT_BEFORE_FORWARD)
_AFTER_EVENTS = (Event.EPOCH_END, Event.BATCH_END, Event.AFTER_DATALOADER, Event.AFTER_TRAIN_BATCH, Event.AFTER_FORWARD,
                 Event.AFTER_LOSS, Event.AFTER_BACKWARD, Event.EVAL_END, Event.EVAL_BATCH_END, Event.EVAL_AFTER_FORWARD,
                 Event.FIT_END, Event.PREDICT_END, Event.PREDICT_BATCH_END, Event.PREDICT_AFTER_FORWARD)



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Profiler to record system level metrics."""

from __future__ import annotations

import threading
import time
from typing import TYPE_CHECKING, Dict, cast

import psutil

from composer.core import Callback

if TYPE_CHECKING:
    from composer.core import State
    from composer.loggers import Logger
    from composer.profiler import Profiler

__all__ = ['SystemProfiler']


class SystemProfiler(Callback):
    """The SystemProfiler records system level metrics.

    .. note::

        The Composer :class:`.Trainer` automatically creates an instance of this
        :class:`.SystemProfiler` callback whenever any of the System Profiler arguments (``sys_prof_cpu``,
        ``sys_prof_memory``, ``sys_prof_disk``, or ``sys_prof_net``) are enabled.

        When using the Composer :class:`.Trainer`, one does not need to directly create an
        instance of this :class:`.SystemProfiler` callback.

    Args:
        profile_cpu (bool): Whether to record cpu statistics (Default: ``True``)
        profile_memory (bool): Whether to record memory statistics (Default: ``False``)
        profile_disk (bool): Whether to record disk I/O statistics (Default: ``False``)
        profile_net (bool): Whether to record network I/O statistics (Default: ``False``)
        stats_thread_interval_seconds (float): Interval to record system-level stats, in seconds. (Default: every ``0.5`` seconds)
    """

    def __init__(self,
                 profile_cpu: bool = True,
                 profile_memory: bool = False,
                 profile_disk: bool = False,
                 profile_net: bool = False,
                 stats_thread_interval_seconds: float = 0.5) -> None:

        self.profile_cpu = profile_cpu
        self.profile_disk = profile_disk
        self.profile_memory = profile_memory
        self.profile_net = profile_net
        self.stats_thread_interval_seconds = stats_thread_interval_seconds
        self.finished_event = threading.Event()

    def init(self, state: State, logger: Logger):
        del logger  # unused
        if state.profiler is None:
            raise RuntimeError(('The Composer Profiler was not enabled, which is required to use the '
                                f'{type(self).__name__}. To enable, set the `prof_schedule` argument of the Trainer.'))

        # Start the stats thread
        self.finished_event.clear()
        threading.Thread(target=self._stats_thread, daemon=True, args=[state.profiler]).start()

    def close(self, state: State, logger: Logger) -> None:
        self.finished_event.set()

    def _stats_thread(self, profiler: Profiler):
        from composer.callbacks import memory_monitor
        """Gathers requested system metrics at :attr:`SystemProfiler.stats_thread_interval_seconds` interval."""

        psutil.disk_io_counters.cache_clear()
        psutil.net_io_counters.cache_clear()
        if self.profile_cpu:
            psutil.cpu_percent()  # spin it once to clear the default 0.0 value on the first call

        while not self.finished_event.is_set():
            if self.profile_cpu:
                cpu_percent = psutil.cpu_percent()
                profiler.marker(name='cpu', categories=['cpu']).counter({'cpu_percent': cpu_percent})

            if self.profile_memory:
                cuda_memory_stats = memory_monitor._get_memory_report()
                for name, val in cuda_memory_stats.items():
                    profiler.marker(f'memory/cuda/{name}', categories=['memory']).counter({name: val})
                swap_memory = psutil.swap_memory()
                profiler.marker('memory/swap', categories=['memory']).counter({
                    'used_gb': swap_memory.used / 2**9,
                    'free_gb': swap_memory.free / 2**9
                })
                virtual_memory = psutil.virtual_memory()
                profiler.marker('memory/virtual', categories=['memory']).counter({
                    'used_gb': virtual_memory.used / 2**9,
                    'available_gb': virtual_memory.available / 2**9
                })

            if self.profile_disk:
                disk_io_counters = cast(Dict[str, psutil._common.sdiskio], psutil.disk_io_counters(perdisk=True))
                for disk_name, disk_stats in disk_io_counters.items():
                    for field_name in ('read_count', 'write_count', 'read_bytes', 'write_bytes', 'read_time',
                                       'write_time', 'busy_time'):
                        profiler.marker(f'disk/{disk_name}/{field_name}',
                                        categories=['disk']).counter({'field_name': getattr(disk_stats, field_name)})

            if self.profile_net:
                net_io_counters = cast(Dict[str, psutil._common.snetio], psutil.net_io_counters(pernic=True))
                for nic, nic_stats in net_io_counters.items():
                    profiler.marker(f'network/{nic}/kb_sent',
                                    categories=['net']).counter({'kb_sent': nic_stats.bytes_sent / 2**3})
                    profiler.marker(f'network/{nic}/kb_recv',
                                    categories=['net']).counter({'kb_recv': nic_stats.bytes_recv / 2**3})

            time.sleep(self.stats_thread_interval_seconds)



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Utility and helper functions for datasets."""

import logging
import textwrap
from typing import Callable, List, Tuple, Union

import numpy as np
import torch
from PIL import Image
from torchvision import transforms
from torchvision.datasets import VisionDataset

from composer.core import Batch

__all__ = [
    'add_vision_dataset_transform',
    'NormalizationFn',
    'pil_image_collate',
]

log = logging.getLogger(__name__)


class NormalizationFn:
    """Normalizes input data and removes the background class from target data if desired.

    An instance of this class can be used as the ``device_transforms`` argument
    when constructing a :class:`~composer.core.data_spec.DataSpec`. When used here,
    the data will normalized after it has been loaded onto the device (i.e., GPU).

    Args:
        mean (Tuple[float, float, float]): The mean pixel value for each channel (RGB) for
            the dataset.
        std (Tuple[float, float, float]): The standard deviation pixel value for each
            channel (RGB) for the dataset.
        ignore_background (bool): If ``True``, ignore the background class in the training
            loss. Only used in semantic segmentation. Default: ``False``.
    """

    def __init__(self,
                 mean: Tuple[float, float, float],
                 std: Tuple[float, float, float],
                 ignore_background: bool = False):
        self.mean = mean
        self.std = std
        self.ignore_background = ignore_background

    def __call__(self, batch: Batch):
        xs, ys = batch
        assert isinstance(xs, torch.Tensor)
        assert isinstance(ys, torch.Tensor)
        device = xs.device

        if not isinstance(self.mean, torch.Tensor):
            self.mean = torch.tensor(self.mean, device=device)
            self.mean = self.mean.view(1, 3, 1, 1)
        if not isinstance(self.std, torch.Tensor):
            self.std = torch.tensor(self.std, device=device)
            self.std = self.std.view(1, 3, 1, 1)

        xs = xs.float()
        xs = xs.sub_(self.mean).div_(self.std)
        if self.ignore_background:
            ys = ys.sub_(1)
        return xs, ys


def pil_image_collate(
        batch: List[Tuple[Image.Image, Union[Image.Image, np.ndarray]]],
        memory_format: torch.memory_format = torch.contiguous_format) -> Tuple[torch.Tensor, torch.Tensor]:
    """Constructs a length 2 tuple of torch.Tensors from datasets that yield samples of type
    :class:`PIL.Image.Image`.

    This function can be used as the ``collate_fn`` argument of a :class:`torch.utils.data.DataLoader`.

    Args:
        batch (List[Tuple[Image.Image, Union[Image.Image, np.ndarray]]]): List of (image, target) tuples
            that will be aggregated and converted into a single (:class:`~torch.Tensor`, :class:`~torch.Tensor`)
            tuple.

        memory_format (torch.memory_format): The memory format for the input and target tensors.

    Returns:
        (torch.Tensor, torch.Tensor): Tuple of (image tensor, target tensor)
            The image tensor will be four-dimensional (NCHW or NHWC, depending on the ``memory_format``).
    """
    imgs = [sample[0] for sample in batch]
    w, h = imgs[0].size
    image_tensor = torch.zeros((len(imgs), 3, h, w), dtype=torch.uint8).contiguous(memory_format=memory_format)

    # Convert targets to torch tensor
    targets = [sample[1] for sample in batch]
    if isinstance(targets[0], Image.Image):
        target_dims = (len(targets), targets[0].size[1], targets[0].size[0])
    else:
        target_dims = (len(targets),)
    target_tensor = torch.zeros(target_dims, dtype=torch.int64).contiguous(memory_format=memory_format)

    for i, img in enumerate(imgs):
        nump_array = np.asarray(img, dtype=np.uint8)
        if nump_array.ndim < 3:
            nump_array = np.expand_dims(nump_array, axis=-1)

        nump_array = np.rollaxis(nump_array, 2).copy()
        if nump_array.shape[0] != 3:
            assert nump_array.shape[0] == 1, 'unexpected shape'
            nump_array = np.resize(nump_array, (3, h, w))
        assert image_tensor.shape[1:] == nump_array.shape, 'shape mismatch'

        image_tensor[i] += torch.from_numpy(nump_array)
        target_tensor[i] += torch.from_numpy(np.array(targets[i], dtype=np.int64))

    return image_tensor, target_tensor


def add_vision_dataset_transform(dataset: VisionDataset, transform: Callable, is_tensor_transform: bool = False):
    """Add a transform to a dataset's collection of transforms.

    Args:
        dataset (VisionDataset): A torchvision dataset.
        transform (Callable): Function to be added to the dataset's collection of
            transforms.
        is_tensor_transform (bool): Whether ``transform`` acts on data of the type
            :class:`~torch.Tensor`. default: ``False``.

            * If ``True``, and :class:`~torchvision.transforms.ToTensor` is present in the transforms of the
              ``dataset``, then ``transform`` will be inserted after the
              :class:`~torchvision.transforms.ToTensor` transform.
            * If ``False`` and :class:`~torchvision.transforms.ToTensor` is present, the ``transform`` will be
              inserted before :class:`~torchvision.transforms.ToTensor`.
            * If :class:`~torchvision.transforms.ToTensor` is not present, the transform will be appended to
              the end of collection of transforms.

    Returns:
        None: The ``dataset`` is modified in-place.
    """

    transform_added_logstring = textwrap.dedent(f"""\
        Transform {transform} added to dataset.
        Dataset now has the following transforms: {dataset.transform}""")

    if dataset.transform is None:
        dataset.transform = transform
        log.warning(transform_added_logstring)
    elif isinstance(dataset.transform, transforms.Compose):
        insertion_index = len(dataset.transform.transforms)
        for i, t in enumerate(dataset.transform.transforms):
            if isinstance(t, transforms.ToTensor):
                insertion_index = i
                break
        if is_tensor_transform:
            insertion_index += 1
        dataset.transform.transforms.insert(insertion_index, transform)
        log.warning(transform_added_logstring)
    else:  # transform is some other basic transform, join using Compose
        if isinstance(dataset.transform, transforms.ToTensor) and not is_tensor_transform:
            dataset.transform = transforms.Compose([transform, dataset.transform])
            log.warning(transform_added_logstring)
        else:
            dataset.transform = transforms.Compose([dataset.transform, transform])
            log.warning(transform_added_logstring)



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Algorithm Passes reorder or modify the execution of algorithms by the Engine.

The order in which algorithms are run matters significantly during composition. For example, the
:class:`.SelectiveBackprop` algorithm runs on the :attr:`.Event.AFTER_DATALOADER` event and must run before
any data augmentations. :class:`.Engine` runs re-ordering passes to resolve such ordering issues or conflicts.

These modifications are represented as algorithm passes, which are functions that modify a list of algorithms.

For example, an algorithm pass that ensures a certain algorithm runs last, would be implemented as:

.. code-block:: python

   def run_last(algorithms: Sequence[Algorithm], event: Event) -> Sequence[Algorithm]:
      algorithms = sorted(algorithms, key=lambda x: isinstance(x, MyAlgorithm))

The passes in this module are registered by default into :class:`.Engine`.
"""
import warnings
from typing import Any, Callable, Sequence, TypeVar

from composer.core.algorithm import Algorithm
from composer.core.event import Event

T = TypeVar('T')

AlgorithmPass = Callable[[Sequence[Algorithm], Event], Sequence[Algorithm]]


def sort_to_front(list_to_sort: Sequence[T], cls: Any) -> Sequence[T]:
    """Helper function to sort instances of a provided class to the front.

    Example:

        .. testsetup::

            from composer.core.passes import sort_to_front

        .. doctest::

            >>> sort_to_front([1, 'b', 3], str)
            ['b', 1, 3]

    Args:
        list_to_sort: list of objects to sort
        cls: sorts all objects of this class to the front

    Returns:
        sorted_list: Sorted List

    """
    return sorted(list_to_sort, key=lambda x: not isinstance(x, cls))


def sort_to_back(list_to_sort: Sequence[T], cls: Any) -> Sequence[T]:
    """Helper function to sort instances of a provided class to the back.

    Example:

        .. testsetup::

            from composer.core.passes import sort_to_back

        .. doctest::

            >>> sort_to_back([1, 'b', 3], str)
            [1, 3, 'b']

    Args:
        list_to_sort: list of objects to sort
        cls: sorts all objects of this class to the back

    Returns:
        sorted_list: Sorted List

    """
    return sorted(list_to_sort, key=lambda x: isinstance(x, cls))


def sort_selective_backprop_first(algorithms: Sequence[Algorithm], event: Event) -> Sequence[Algorithm]:
    """Selective Backprop should run before any algorithms modify the loss.

    :class:`.SelectiveBackprop` runs after the dataloader returns the batch and executes an extra forward pass to rank
    and prune the examples in the batch by loss. To ensure a clean estimate of loss, :class:`.SelectiveBackprop` should
    run before any other data augmentations (e.g., :class:`.MixUp`) on the :attr:`.Event.AFTER_DATALOADER` event.

    """
    from composer.algorithms import SelectiveBackprop
    return sort_to_front(algorithms, cls=SelectiveBackprop)


def sort_fused_layernorm_last(algorithms: Sequence[Algorithm], event: Event) -> Sequence[Algorithm]:  #noqa: D403
    """FusedLayerNorm should run after other algorithms that add LayerNorms (e.g. GatedLinearUnits).

    This ensures that all LayerNorms are converted to optimized fused versions.

    """
    from composer.algorithms import FusedLayerNorm
    return sort_to_back(algorithms, cls=FusedLayerNorm)


def sort_low_precision_layernorm_last(algorithms: Sequence[Algorithm],
                                      event: Event) -> Sequence[Algorithm]:  #noqa: D403
    """LowPrecisionLayerNorm should run after other algorithms that add LayerNorms (e.g. GatedLinearUnits).

    This ensures that all LayerNorms are converted to the intended precision.

    """
    from composer.algorithms import LowPrecisionLayerNorm
    return sort_to_back(algorithms, cls=LowPrecisionLayerNorm)


def set_filo_order(algorithms: Sequence[Algorithm], event: Event) -> Sequence[Algorithm]:
    """Establish a FILO order of algorithms ``before_`` and ``after_`` events.

    For the events that follow the ``before_*`` and ``after_*`` pattern (e.g., :attr:`.Event.BEFORE_LOSS`
    and :attr:`.Event.AFTER_LOSS), the ordering of algorithms is reversed for the ``after_*`` events.
    For example, four given algorithms ``A``, ``B``, ``C``, and ``D`` will run in ``ABCD`` ordering on
    the ``before_*`` event while ``DCBA`` ordering on the ``after_*`` event.

    This allows algorithms to "clean up" their changes. For example, :class:`.LabelSmoothing` will smooth the labels
    upon the :attr:`.Event.BEFORE_LOSS` event and then restore the original unsmoothed labels on the
    :attr:`.Event.AFTER_LOSS` event.

    Events with the pattern ``_start`` or ``_end`` will not be affected.
    """
    if event.name.startswith('AFTER_'):
        return list(reversed(algorithms))

    return algorithms


def warn_if_multiple_loss_interpolation(algorithms: Sequence[Algorithm], event: Event) -> Sequence[Algorithm]:
    """Multiple algorithms that interpolate the loss may have unexpected behavior."""
    from composer.algorithms.warnings import NotIntendedUseWarning

    is_interpolate = [a for a in algorithms if hasattr(a, 'interpolate_loss') and a.interpolate_loss]  # type: ignore
    if len(is_interpolate) > 1:
        warnings.warn(
            NotIntendedUseWarning(
                f'Multiple algorithms interpolating the loss can lead to unexpected behavior: {is_interpolate}'))

    return algorithms



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Core SelectiveBackprop class and functions."""

from __future__ import annotations

import inspect
from typing import Any, Callable, Optional, Sequence, Tuple, Union

import numpy as np
import torch
from torch.nn import functional as F

from composer.core import Algorithm, Event, State, get_precision_context
from composer.loggers import Logger
from composer.models import ComposerModel

__all__ = ['SelectiveBackprop', 'select_using_loss', 'should_selective_backprop']


def should_selective_backprop(
    current_duration: float,
    batch_idx: int,
    start: float = 0.5,
    end: float = 0.9,
    interrupt: int = 2,
) -> bool:
    """Decides if selective backprop should be run based on time in training.

    Returns true if the ``current_duration`` is between ``start`` and
    ``end``. It is recommended that SB be applied during the later stages of
    a training run, once the model has already "learned" easy examples.

    To preserve convergence, SB can be interrupted with vanilla minibatch
    gradient steps every ``interrupt`` steps. When ``interrupt=0``, SB will be
    used at every step during the SB interval. When ``interrupt=2``, SB will
    alternate with vanilla minibatch steps.

    Args:
        current_duration (float): The elapsed training duration. Must be
            within ``[0.0, 1.0)``.
        batch_idx (int): The current batch within the epoch.
        start (float, optional): The duration at which selective backprop
            should be enabled, as a percentage. Default: ``0.5``.
        end (float, optional): The duration at which selective backprop
            should be disabled. Default: ``0.9``.
        interrupt (int, optional): The number of batches between vanilla
            minibatch gradient updates. Default: ``2``.

    Returns:
        bool: If selective backprop should be performed on this batch.
    """
    is_interval = ((current_duration >= start) and (current_duration < end))
    is_step = ((interrupt == 0) or ((batch_idx + 1) % interrupt != 0))

    return is_interval and is_step


def select_using_loss(input: torch.Tensor,
                      target: torch.Tensor,
                      model: Callable[[Union[torch.Tensor, Sequence[torch.Tensor]]], torch.Tensor],
                      loss_fun: Callable,
                      keep: float = 0.5,
                      scale_factor: float = 1) -> Tuple[torch.Tensor, torch.Tensor]:
    """Prunes minibatches as a subroutine of :class:`.SelectiveBackprop`. Computes the loss function on the provided training
    examples and runs minibatches according to the difficulty. The fraction of the minibatch that is kept for gradient
    computation is specified by the argument ``0 <= keep <= 1``.

    To speed up SB's selection forward pass, the argument ``scale_factor`` can
    be used to spatially downsample input tensors. The full-sized inputs
    will still be used for the weight gradient computation.

    Args:
        input (torch.Tensor): Input tensor to prune.
        target (torch.Tensor): Output tensor to prune.
        model (Callable): Model with which to predict outputs.
        loss_fun (Callable): Loss function of the form ``loss(outputs, targets, reduction='none')``.
            The function must take the keyword argument ``reduction='none'``
            to ensure that per-sample losses are returned.
        keep (float, optional): Fraction of examples in the batch to keep. Default: ``0.5``.
        scale_factor (float, optional): Multiplier between 0 and 1 for spatial size. Downsampling
            requires the input tensor to be at least 3D. Default: ``1``.

    Returns:
        (torch.Tensor, torch.Tensor): The pruned batch of inputs and targets

    Raises:
        ValueError: If ``scale_factor > 1``.
        TypeError: If ``loss_fun > 1`` has the wrong signature or is not callable.

    .. note::

        This function runs an extra forward pass through the model on the batch of data.
        If you are using a non-default precision, ensure that this forward pass
        runs in your desired precision. For example:

    .. testsetup::

        N_sb, D_sb = 16, 8
        X_sb, y_sb = torch.randn(N_sb, D_sb), torch.randint(2, (N_sb,))
        lin_model = torch.nn.Linear(X_sb.shape[1], 1)

    .. doctest::

        >>> import torch
        >>> from composer.algorithms.selective_backprop import select_using_loss
        >>> with torch.cuda.amp.autocast(True):
        ...     X_new, y_new = select_using_loss(
        ...         X_sb,
        ...         y_sb,
        ...         lin_model,
        ...         loss_fun,
        ...         keep=0.5,
        ...         scale_factor=1
        ...     )
    """
    INTERPOLATE_MODES = {3: 'linear', 4: 'bilinear', 5: 'trilinear'}

    interp_mode = 'bilinear'
    if scale_factor != 1:
        if input.dim() not in INTERPOLATE_MODES:
            raise ValueError(f'Input must be 3D, 4D, or 5D if scale_factor != 1, got {input.dim()}')
        interp_mode = INTERPOLATE_MODES[input.dim()]

    if scale_factor > 1:
        raise ValueError('scale_factor must be <= 1')

    if callable(loss_fun):
        sig = inspect.signature(loss_fun)
        if not 'reduction' in sig.parameters:
            raise TypeError('Loss function `loss_fun` must take a keyword argument `reduction`.')
    else:
        raise TypeError('Loss function must be callable')

    with torch.no_grad():
        N = input.shape[0]

        # Maybe interpolate
        if scale_factor < 1:
            X_scaled = F.interpolate(input,
                                     scale_factor=scale_factor,
                                     mode=interp_mode,
                                     align_corners=False,
                                     recompute_scale_factor=False)
        else:
            X_scaled = input

        # Get per-examples losses
        out = model(X_scaled)
        losses = loss_fun(out, target, reduction='none')

        # Sort losses
        sorted_idx = torch.argsort(losses)
        n_select = int(keep * N)

        # Sample by loss
        percs = np.arange(0.5, N, 1) / N
        probs = percs**((1.0 / keep) - 1.0)
        probs = probs / np.sum(probs)
        select_percs_idx = np.random.choice(N, n_select, replace=False, p=probs)
        select_idx = sorted_idx[select_percs_idx]

    return input[select_idx], target[select_idx]


class SelectiveBackprop(Algorithm):
    """Selectively backpropagate gradients from a subset of each batch.

    Based on (`Jiang et al, 2019`_), Selective Backprop (SB) prunes minibatches
    according to the difficulty of the individual training examples, and only
    computes weight gradients over the pruned subset, reducing iteration time, and
    speeding up training.

    The fraction of the minibatch that is kept for gradient computation is
    specified by the argument ``0 <= keep <= 1``.

    To speed up SB's selection forward pass, the argument ``scale_factor`` can
    be used to spatially downsample input image tensors. The full-sized inputs
    will still be used for the weight gradient computation.

    To preserve convergence, SB can be interrupted with vanilla minibatch
    gradient steps every ``interrupt`` steps. When ``interrupt=0``, SB will be
    used at every step during the SB interval. When ``interrupt=2``, SB will
    alternate with vanilla minibatch steps.

    .. _Jiang et al, 2019: https://arxiv.org/abs/1910.00762

    Args:
        start (float, optional): SB interval start as fraction of training duration.
            Default: ``0.5``.
        end (float, optional): SB interval end as fraction of training duration.
            Default: ``0.9``.
        keep (float, optional): fraction of minibatch to select and keep for gradient computation.
            Default: ``0.5``.
        scale_factor (float, optional): scale for downsampling input for selection forward pass.
            Default: ``1.``.
        interrupt (int, optional): interrupt SB with a vanilla minibatch step every
            ``interrupt`` batches. Default: ``2``.
        input_key (str | int | Tuple[Callable, Callable] | Any, optional): A key that indexes to the input
            from the batch. Can also be a pair of get and set functions, where the getter
            is assumed to be first in the pair.  The default is 0, which corresponds to any sequence, where the first element
            is the input. Default: ``0``.
        target_key (str | int | Tuple[Callable, Callable] | Any, optional): A key that indexes to the target
            from the batch. Can also be a pair of get and set functions, where the getter
            is assumed to be first in the pair. The default is 1, which corresponds to any sequence, where the second element
            is the target. Default: ``1``.

    Example:
        .. testcode::

            from composer.algorithms import SelectiveBackprop
            algorithm = SelectiveBackprop(start=0.5, end=0.9, keep=0.5)
            trainer = Trainer(
                model=model,
                train_dataloader=train_dataloader,
                eval_dataloader=eval_dataloader,
                max_duration="1ep",
                algorithms=[algorithm],
                optimizers=[optimizer]
            )
    """

    def __init__(
        self,
        start: float = 0.5,
        end: float = 0.9,
        keep: float = 0.5,
        scale_factor: float = 1.,
        interrupt: int = 2,
        input_key: Union[str, int, Tuple[Callable, Callable], Any] = 0,
        target_key: Union[str, int, Tuple[Callable, Callable], Any] = 1,
    ):
        self.start = start
        self.end = end
        self.keep = keep
        self.scale_factor = scale_factor
        self.interrupt = interrupt
        self._loss_fn = None  # set on Event.INIT
        self.input_key, self.target_key = input_key, target_key

    def match(self, event: Event, state: State) -> bool:
        if event == Event.INIT:
            return True
        if event != Event.AFTER_DATALOADER:
            return False

        is_keep = (self.keep < 1)
        if not is_keep:
            return False

        elapsed_duration = state.get_elapsed_duration()
        assert elapsed_duration is not None, 'elapsed duration should be set on Event.AFTER_DATALOADER'

        is_chosen = should_selective_backprop(
            current_duration=float(elapsed_duration),
            batch_idx=int(state.timestamp.batch_in_epoch),
            start=self.start,
            end=self.end,
            interrupt=self.interrupt,
        )
        return is_chosen

    def apply(self, event: Event, state: State, logger: Optional[Logger] = None) -> None:
        if event == Event.INIT:
            if self._loss_fn is None:
                if not isinstance(state.model, ComposerModel):
                    raise RuntimeError('Model must be of type ComposerModel')
                self._loss_fn = state.model.loss
            return
        input, target = state.batch_get_item(key=self.input_key), state.batch_get_item(key=self.target_key)
        assert isinstance(input, torch.Tensor) and isinstance(target, torch.Tensor), \
            'Multiple tensors not supported for this method yet.'

        # Model expected to only take in input, not the full batch
        model = lambda X: state.model((X, None))

        def loss(p, y, reduction='none'):
            assert self._loss_fn is not None, 'loss_fn should be set on Event.INIT'
            return self._loss_fn(p, (torch.Tensor(), y), reduction=reduction)

        with get_precision_context(state.precision):
            new_input, new_target = select_using_loss(input, target, model, loss, self.keep, self.scale_factor)
        state.batch_set_item(self.input_key, new_input)
        state.batch_set_item(self.target_key, new_target)



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Changes the memory format of the model to ``torch.channels_last``.

This usually improves GPU utilization. See the :doc:`Method Card </method_cards/channels_last>` for more details.
"""

from composer.algorithms.channels_last.channels_last import ChannelsLast as ChannelsLast
from composer.algorithms.channels_last.channels_last import apply_channels_last as apply_channels_last

__all__ = ['ChannelsLast', 'apply_channels_last']



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Helpers to gather system information for debugging and bug reporting.

Leverages PyTorch's :mod:`torch.utils.collect_env` package to gather pertinent system information.
The following information is additionally collected to faciliate Comopser specific debug:

* Composer version
* Number of nodes
* Host processor model name
* Host processor physical core count
* Number of accelerators per node
* Accelerator model name

This package can be invoked as a standalone console script or can be invoked from within
an application to gather and generate a system environment report.

The module can be invoked by using the entrypoint alias:

.. code-block::

    $ composer_collect_env

Or manually as a standalone script:

.. code-block::

    $ python composer/utils/collect_env.py

To generate a system report from within a user application see :func:`print_env`.

A custom excepthook wrapper is also provided which extends the original :func:`sys.excepthook`
to automatically collect system information when an exception is raised.

To override the original :func:`sys.excepthook` see :func:`configure_excepthook`.

By default, the Composer custom ``excepthook`` automatically generates the environment report.
To disable automatic environment report generation, use the :func:`disable_env_report` helper
function.  Report generation can be re-enabled by using the :func:`enable_env_report` function.
"""

import functools
import json
import sys
import time
from typing import NamedTuple, Optional, TextIO

import cpuinfo
import importlib_metadata
import psutil

from composer.utils.misc import is_notebook

__all__ = ['configure_excepthook', 'disable_env_report', 'enable_env_report', 'print_env', 'get_composer_env_dict']

# Check if PyTorch is installed
try:
    import torch.utils.collect_env as torchenv
    from torch.cuda import device_count as cuda_device_count
    from torch.cuda import get_device_name as accel_device_name
    from torch.cuda import is_available as cuda_available
    TORCH_AVAILABLE = True
except (ImportError,):
    TORCH_AVAILABLE = False

# Check if Composer is installed
try:
    import composer
    from composer.utils import dist
    COMPOSER_AVAILABLE = True
except (ImportError,):
    COMPOSER_AVAILABLE = False

# Check if we're running in a notebook
IPYTHON_AVAILABLE = is_notebook()
if IPYTHON_AVAILABLE:
    from composer.utils.import_helpers import import_object
    get_ipython = import_object('IPython:get_ipython')
    nb = get_ipython()

# Place to keep track of the original excepthook
_orig_excepthook = None

# Track if excepthook was previously registered, needed for indempotency
_EXCEPTHOOK_REGISTERED = False

# Track if environment report generation on exception is enabled, enabled by default
_ENV_EXCEPTION_REPORT = True


# Same convention as Torch collect_env, create a namedtuple to track collected fields
class ComposerEnv(NamedTuple):
    composer_version: str
    composer_commit_hash: Optional[str]
    node_world_size: int
    host_processor_model_name: str
    host_processor_core_count: int
    local_world_size: int
    accelerator_model_name: str
    cuda_device_count: int


def get_composer_commit_hash() -> Optional[str]:
    # Use PEP-610 to get the commit hash
    # See https://packaging.python.org/en/latest/specifications/direct-url/
    # Try both package names that Composer is released under
    try:
        files = importlib_metadata.files('mosaicml')
    except importlib_metadata.PackageNotFoundError:
        try:
            files = importlib_metadata.files('composer')
        except importlib_metadata.PackageNotFoundError:
            return

    if files is None:
        return
    files = [f for f in files if str(f).endswith('direct_url.json')]
    if len(files) == 0:
        return
    f = files[0]
    direct_url = json.loads(f.read_text())
    vcs_info = direct_url.get('vcs_info', {})
    commit_id = vcs_info.get('commit_id')
    return commit_id


# Helper functions to get Composer environment information
def get_composer_version() -> str:
    """Query the Composer version."""
    return str(composer.__version__)


@functools.lru_cache(maxsize=1)
def get_host_processor_name() -> str:
    """Query the host processor name."""
    cpu_info = cpuinfo.get_cpu_info()
    return str(cpu_info.get('brand_raw', 'CPU'))


def get_host_processor_cores() -> int:
    """Determines the number of physical host processor cores."""
    return psutil.cpu_count(logical=False)


def get_node_world_size() -> int:
    """Query the number of nodes."""
    return int(dist.get_world_size() / dist.get_local_world_size())


def get_accel_model_name() -> str:
    """Query the accelerator name."""
    return accel_device_name(None) if cuda_available() else 'N/A'


def get_local_world_size() -> int:
    """Determines the number of accelerators per node."""
    return dist.get_local_world_size() if cuda_available() else 0


def get_cuda_device_count() -> int:
    """Get the number of CUDA devices on the system."""
    return cuda_device_count() if TORCH_AVAILABLE else 0


# Exception message and environment report
COMPOSER_OPEN_ISSUE_URL = 'https://github.com/mosaicml/composer/issues/new/choose'


def _exc_report(exc_type) -> None:
    """Produces exception report (exception message + environment report).

    Args:
        exc_type (Exception): Type of exception.
    """
    EXCEPTION_MSG = f'Bugs can be reported at: {COMPOSER_OPEN_ISSUE_URL}\n'

    # Don't print exception report for KeyboardInterrupt
    if not issubclass(exc_type, KeyboardInterrupt):
        if issubclass(exc_type, AssertionError):
            EXCEPTION_SEV_MSG = f'This is a Composer bug. Please submit a bug report.\n'
        elif issubclass(exc_type, RuntimeError):
            EXCEPTION_SEV_MSG = f'This could be due to user error but is most likely a Composer bug.\n'
        elif issubclass(exc_type, ValueError) or issubclass(exc_type, TypeError):
            EXCEPTION_SEV_MSG = f'This was most likely due to user error but please submit a bug report if you suspect a Composer issue.\n'
        else:
            EXCEPTION_SEV_MSG = f'If you would like support debugging, submit a bug report or reach out to us on our community channels.\n'

        print('\n-------------------\n' + EXCEPTION_SEV_MSG + EXCEPTION_MSG, file=sys.stderr)

        # Only print environment report if enabled
        global _ENV_EXCEPTION_REPORT
        if _ENV_EXCEPTION_REPORT:
            print('Please include details on how to reproduce the issue and attach the following environment report:\n')
            print_env(sys.stderr)
        else:
            print(
                "Please run the \'composer_collect_env\' utility and include your environment information with the bug report\n"
            )


def enable_env_report() -> None:
    """Enable environment report generation on exception."""
    global _ENV_EXCEPTION_REPORT
    _ENV_EXCEPTION_REPORT = True


def disable_env_report() -> None:
    """Disable environment report generation on exception."""
    global _ENV_EXCEPTION_REPORT
    _ENV_EXCEPTION_REPORT = False


# Excepthook wrapper, wraps default excepthook and prints env info
def _custom_exception_handler(type, value, tb) -> None:
    """Custom exception wrapper for sys.excepthook."""
    _exc_report(exc_type=type)
    assert _orig_excepthook
    _orig_excepthook(type, value, tb)


# Custom exception handler for IPython notebooks
def _nb_custom_exception_handler(self, type, value, tb, tb_offset=None):
    """Custom exception handler for IPython."""
    _exc_report(exc_type=type)
    self.showtraceback((type, value, tb), tb_offset=tb_offset)  # standard IPython's printout


# Public function to register excethook wrapper
def configure_excepthook() -> None:
    """Collect and print system information when :func:`sys.excepthook` is called.

    The custom exception handler causes an exception message to be printed when :func:`sys.excepthook`
    is called.  The exception message provides the user with information on the nature of the exception
    and directs the user to file GitHub issues as appropriate.

    By default, the custom exception handler also generates an environment report users can attach to
    bug reports.  Environment report generation can be optionally enabled/disabled by using the
    :func:`enable_env_report` and :func:`disable_env_report` helper functions, respectively.

    Additioanlly, the custom exceptionhook checks if the user is running from an IPython session and
    sets up the custom exception handler accordingly.

    To override the default :func:`sys.excepthook` with the custom except hook:

    .. testsetup:: composer.utils.collect_env.configure_excepthook

        import sys
        from composer.utils.collect_env import configure_excepthook
        import composer.utils.collect_env as ce

        sys.excepthook = sys.__excepthook__
        ce._EXCEPTHOOK_REGISTERED = False

    .. doctest:: composer.utils.collect_env.configure_excepthook

        >>> configure_excepthook()
        >>> sys.excepthook
        <function _custom_exception_handler at ...>
    """
    global _EXCEPTHOOK_REGISTERED
    # Needs to be indempotent across multiple trainers, don't register if we've already registered
    if not _EXCEPTHOOK_REGISTERED:
        # Custom exceptions work differntly in notebooks
        if IPYTHON_AVAILABLE:
            # Set custom handler on Exception base class to apply to all exceptions
            nb.set_custom_exc((Exception,), _nb_custom_exception_handler)
        else:
            # Save original excepthook and override
            global _orig_excepthook
            _orig_excepthook = sys.excepthook
            sys.excepthook = _custom_exception_handler

        _EXCEPTHOOK_REGISTERED = True


# Get Torch environment info
def get_torch_env() -> str:
    """Query Torch system environment via :mod:`torch.utils.collect_env`."""
    return torchenv.get_pretty_env_info()


# Composer environment information string output format
_COMPOSER_ENV_INFO_FORMAT = """
Composer version: {composer_version}
Composer commit hash: {composer_commit_hash}
Host processor model name: {host_processor_model_name}
Host processor core count: {host_processor_core_count}
Number of nodes: {node_world_size}
Accelerator model name: {accelerator_model_name}
Accelerators per node: {local_world_size}
CUDA Device Count: {cuda_device_count}
""".strip()


# Get composer environment info as a dictionary
def get_composer_env_dict() -> dict:
    """Query Composer pertinent system information as a dict."""
    mutable_dict = ComposerEnv(
        composer_version=get_composer_version(),
        composer_commit_hash=get_composer_commit_hash(),
        host_processor_model_name=get_host_processor_name(),
        host_processor_core_count=get_host_processor_cores(),
        node_world_size=get_node_world_size(),
        accelerator_model_name=get_accel_model_name(),
        local_world_size=get_local_world_size(),
        cuda_device_count=get_cuda_device_count(),
    )._asdict()
    return mutable_dict


# Get Composer environment info
def get_composer_env() -> str:
    """Query Composer pertinent system information."""
    mutable_dict = get_composer_env_dict()
    return _COMPOSER_ENV_INFO_FORMAT.format(**mutable_dict)


# Generate and print environment report
def print_env(file: Optional[TextIO] = None) -> None:
    """Generate system information report.

    Example:
    .. code-block:: python

        from composer.utils.collect_env import print_env

        print_env()

    Sample Report:

    .. code-block:: text

        ---------------------------------
        System Environment Report
        Created: 2022-04-27 00:25:33 UTC
        ---------------------------------

        PyTorch information
        -------------------
        PyTorch version: 1-91+cu111
        Is debug build: False
        CUDA used to build PyTorch: 111
        ROCM used to build PyTorch: N/A

        OS: Ubuntu 18.04.6 LTS (x86_64)
        GCC version: (Ubuntu 7.5.0-3ubuntu1~18.04) 7.5.0
        Clang version: Could not collect
        CMake version: version 3.10.2
        Libc version: glibc-2.27

        Python version: 3.8 (64-bit runtime)
        Python platform: Linux-5.8.0-63-generic-x86_64-with-glibc2.27
        Is CUDA available: True
        CUDA runtime version: 11.1.105
        GPU models and configuration:
        GPU 0: NVIDIA GeForce RTX 3080
        GPU 1: NVIDIA GeForce RTX 3080
        GPU 2: NVIDIA GeForce RTX 3080
        GPU 3: NVIDIA GeForce RTX 3080

        Nvidia driver version: 470.57.02
        cuDNN version: Probably one of the following:
        /usr/lib/x86_64-linux-gnu/libcudnn.so.8.0.5
        /usr/lib/x86_64-linux-gnu/libcudnn_adv_infer.so.8.0.5
        /usr/lib/x86_64-linux-gnu/libcudnn_adv_train.so.8.0.5
        /usr/lib/x86_64-linux-gnu/libcudnn_cnn_infer.so.8.0.5
        /usr/lib/x86_64-linux-gnu/libcudnn_cnn_train.so.8.0.5
        /usr/lib/x86_64-linux-gnu/libcudnn_ops_infer.so.8.0.5
        /usr/lib/x86_64-linux-gnu/libcudnn_ops_train.so.8.0.5
        HIP runtime version: N/A
        MIOpen runtime version: N/A

        Versions of relevant libraries:
        [pip3] numpy==1.22.3
        [pip3] pytorch-ranger==0.1.1
        [pip3] torch==1.9.1+cu111
        [pip3] torch-optimizer==0.1.0
        [pip3] torchmetrics==0.7.3
        [pip3] torchvision==0.10.1+cu111
        [pip3] vit-pytorch==0.27.0
        [conda] Could not collect


        Composer information
        --------------------
        Composer version: 0.8.2
        Composer commit hash: 9e14a47562def0baa414242c36954eb3083dcd46
        Host processor model name: AMD EPYC 7502 32-Core Processor
        Host processor core count: 64
        Number of nodes: 1
        Accelerator model name: NVIDIA GeForce RTX 3080
        Accelerators per node: 1
        CUDA Device Count: 4


    Args:
        file (TextIO, optional): File handle, `sys.stdout` or `sys.stderr`. Defaults to `sys.stdout`.
    """
    # Set stdout during runtime if no output file is specified
    if file is None:
        file = sys.stdout

    # Creation timestamp for report
    creation_time = time.strftime('%Y-%m-%d %H:%M:%S %Z', time.localtime(time.time()))
    report_hdr = f'---------------------------------\n' + \
                 f'System Environment Report        \n' + \
                 f'Created: {creation_time}\n'          + \
                 f'---------------------------------\n'

    # Torch section
    print(report_hdr, file=file)
    print('PyTorch information', file=file)
    print('-------------------', file=file)
    if TORCH_AVAILABLE:
        # Only print Torch system info if installed
        print(get_torch_env() + '\n\n', file=file)
    else:
        print('Torch not installed', file=file)

    print('Composer information', file=file)
    print('--------------------', file=file)
    if COMPOSER_AVAILABLE:
        # Only print Composer system info if installed
        print(get_composer_env() + '\n\n', file=file)
    else:
        print('Composer not installed', file=file)


# Invoke as standalone CLI script
def main() -> None:
    print('Collecting system information...')
    print_env()


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""Base class for algorithms that improve a model's quality or efficiency."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Dict, Optional

from composer.core.serializable import Serializable

if TYPE_CHECKING:
    from composer.core import Event, State
    from composer.loggers import Logger

__all__ = ['Algorithm']


class Algorithm(Serializable, ABC):
    """Base class for algorithms.

    Algorithms are pieces of code which run at specific events (see :class:`.Event`) in the training loop.
    Algorithms modify the trainer's :class:`.State`, generally with the effect of improving the model's quality
    or increasing the efficiency and throughput of the training loop.

    Algorithms must implement the following two methods:
      +----------------+-------------------------------------------------------------------------------+
      | Method         | Description                                                                   |
      +================+===============================================================================+
      | :func:`match`  | returns whether the algorithm should be run given the current                 |
      |                | :class:`.Event` and :class:`.State`.                                          |
      +----------------+-------------------------------------------------------------------------------+
      | :func:`apply`  | Executes the algorithm's code and makes an in-place change                    |
      |                | to the :class:`.State`.                                                       |
      +----------------+-------------------------------------------------------------------------------+
    """

    def __init__(self, *args, **kwargs):  # Stub signature for PyRight
        del args, kwargs  # unused
        pass

    @property
    def find_unused_parameters(self) -> bool:
        """Indicates whether this algorithm may cause some model parameters to be unused. Defaults to False.

        For example, it is used to tell :class:`torch.nn.parallel.DistributedDataParallel` (DDP) that some parameters
        will be frozen during training, and hence it should not expect gradients from them. All algorithms which do any
        kind of parameter freezing should override this function to return ``True``.

        .. note::

           DeepSpeed integration with this function returning True is not tested. It may not work as expected.
        """
        return False

    @property
    def backwards_create_graph(self) -> bool:
        """Return ``True`` to indicate this algorithm requires a second derivative to be computed. Defaults to ``False``.

        If it returns ``True``, ``create_graph=True`` will be passed to :meth:`torch.Tensor.backward` which will result in
        the graph of the gradient also being constructed. This allows the computation of second order derivatives.
        """
        return False

    @staticmethod
    def required_on_load() -> bool:
        """Return `True` to indicate this algorithm is required when loading from a checkpoint which used it."""
        return False

    def state_dict(self) -> Dict[str, Any]:
        return {'repr': self.__repr__()}

    @abstractmethod
    def match(self, event: Event, state: State) -> bool:
        """Determines whether this algorithm should run given the current :class:`.Event` and :class:`.State`.

        Examples:
        To only run on a specific event (e.g., on :attr:`.Event.BEFORE_LOSS`), override match as shown below:

        >>> class MyAlgorithm:
        ...     def match(self, event, state):
        ...         return event == Event.BEFORE_LOSS
        >>> MyAlgorithm().match(Event.BEFORE_LOSS, state)
        True

        To run based on some value of a :class:`.State` attribute, override match as shown below:

        >>> class MyAlgorithm:
        ...     def match(self, event, state):
        ...        return state.timestamp.epoch > 30
        >>> MyAlgorithm().match(Event.BEFORE_LOSS, state)
        False

        See :class:`.State` for accessible attributes.

        Args:
            event (Event): The current event.
            state (State): The current state.

        Returns:
            bool: True if this algorithm should run now.
        """
        raise NotImplementedError(f'implement match() required for {self.__class__.__name__}')

    @abstractmethod
    def apply(self, event: Event, state: State, logger: Logger) -> Optional[int]:
        """Applies the algorithm to make an in-place change to the :class:`.State`.

        Can optionally return an exit code to be stored in a :class:`.Trace`.
        This exit code is made accessible for debugging.

        Args:
            event (Event): The current event.
            state (State): The current state.
            logger (Logger): A logger to use for logging algorithm-specific metrics.

        Returns:
            int or None: exit code that will be stored in :class:`.Trace` and made accessible for debugging.
        """
        raise NotImplementedError(f'implement apply() required for {self.__class__.__name__}')



# Copyright 2022 MosaicML Composer authors
# SPDX-License-Identifier: Apache-2.0

"""The `BERT <https://huggingface.co/docs/transformers/master/en/model_doc/bert>`_ model family using `Hugging Face
Transformers <https://huggingface.co/transformers/>`_."""

from composer.models.bert.model import create_bert_classification as create_bert_classification
from composer.models.bert.model import create_bert_mlm as create_bert_mlm

__all__ = ['create_bert_classification', 'create_bert_mlm']
